#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GPTQ quantization."""

import gc
import math
import re
import time
from functools import partial

import torch
import torch.nn as nn
from tqdm import tqdm

from neural_compressor.torch.algorithms.layer_wise import get_path, load_value, set_module_tensor_to_device
from neural_compressor.torch.utils import (
    get_accelerator,
    get_model_device,
    get_used_cpu_mem_MB,
    is_transformers_imported,
    logger,
    set_module,
)
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

from .modules import INCWeightOnlyLinear

if is_transformers_imported():
    import transformers

    SUPPORTED_LAYERS = [nn.Conv2d, nn.Conv1d, nn.Linear, transformers.Conv1D]
else:
    SUPPORTED_LAYERS = [nn.Conv2d, nn.Conv1d, nn.Linear]
DEBUG = False
accelerator = auto_detect_accelerator()


# ==============model structure related==============
def is_leaf(module):
    """Judge whether a module has no child-modules.

    Args:
        module: torch.nn.Module

    Returns:
        a bool: whether a module has no child-modules.
    """
    children_cnt = 0
    for n in module.children():
        children_cnt += 1
    return True if children_cnt == 0 else False


def trace_gptq_target_blocks(module, module_types=[torch.nn.ModuleList, torch.nn.Sequential]):
    """Search transformer stacked structures, which is critical in LLMs and GPTQ execution.

    Args:
        module: torch.nn.Module
        module_types: List of torch.nn.Module.

    Returns:
        gptq_related_blocks = {
            "embeddings": {}, # Dict embedding layers before transformer stack module,
            "transformers_pre": {}, # TODO
            "transformers_name": string. LLMs' transformer stack module name ,
            "transformers": torch.nn.ModuleList. LLMs' transformer stack module,
            "transformers": {}, Dict# TODO
        }
    """
    find_transformers = False
    if type(module).__name__ == "MixFormerSequentialForCausalLM":  # pragma: no cover
        gptq_related_blocks = {
            "embeddings": {},
            "transformers_pre": {},  # todo
            "transformers_name": "",  # None
            "transformers": [],  # None
            "transformers_post": {},  # todo
        }
        for n, m in module.named_modules():
            if type(m) in module_types:
                gptq_related_blocks["transformers_name"] = n
                gptq_related_blocks["transformers"] = m
                break
            else:
                continue
        for n, m in gptq_related_blocks["transformers"][0].named_modules():
            if is_leaf(m):
                gptq_related_blocks["embeddings"][n] = m
        gptq_related_blocks["transformers"] = gptq_related_blocks["transformers"][1:-1]
    else:
        gptq_related_blocks = {
            "embeddings": {},
            "transformers_pre": {},  # todo
            "transformers_name": "",  # None
            "transformers": [],  # None
            "transformers_post": {},  # todo
        }
        for n, m in module.named_modules():
            if type(m) in module_types:
                # find the block
                gptq_related_blocks["transformers_name"] = n
                gptq_related_blocks["transformers"] = m
                find_transformers = True
                # return gptq_related_blocks
            elif (is_leaf(m) and not find_transformers) or "Embedding" in type(m).__name__:
                # "Embedding" in type(m).__name__ to resolve 'LlamaRotaryEmbedding'
                gptq_related_blocks["embeddings"][n] = m
            elif n.find(gptq_related_blocks["transformers_name"]) == -1 and find_transformers:
                # no longer belong to transformers
                gptq_related_blocks["transformers_post"]["name"] = n
                gptq_related_blocks["transformers_post"]["layer"] = m
            else:
                continue
    return gptq_related_blocks


def find_all_layers(module, name=""):
    """Get all layers."""
    if len(list(module.named_children())) == 0:
        return {name: module}
    res = {}
    for name1, child in module.named_children():
        res.update(find_all_layers(child, name=name + "." + name1 if name != "" else name1))
    return res


def find_layers(module, layers=SUPPORTED_LAYERS, name=""):
    """Get all layers with target types."""
    if type(module) in layers:
        return {name: module}
    else:
        # use string type to find name:
        if isinstance(module, tuple(layers)):
            return {name: module}
        else:
            pass
    res = {}
    for name1, child in module.named_children():
        res.update(find_layers(child, layers=layers, name=name + "." + name1 if name != "" else name1))
    return res


def find_layers_name(module, layers=SUPPORTED_LAYERS, name=""):
    """Get all layers with target types."""
    if type(module) in layers:
        return [name]
    res = []
    for name1, child in module.named_children():
        res += find_layers_name(child, layers=layers, name=name + "." + name1 if name != "" else name1)
    return res


def log_quantizable_layers_per_transformer(transformer_blocks, layers=SUPPORTED_LAYERS):
    """Print all layers which will be quantized in GPTQ algorithm."""
    logger.info("* * Layer to be quantized * *")
    quantizable_layers = []
    for block_id in range(len(transformer_blocks["transformers"])):
        transformer_block = transformer_blocks["transformers"][block_id]
        layers_for_this_tblock = find_layers_name(transformer_block)
        layer_names = [
            (transformer_blocks["transformers_name"] + "." + str(block_id) + "." + layer_name)
            for layer_name in layers_for_this_tblock
        ]
        for name in layer_names:
            logger.info(name)
            quantizable_layers.append(name)
    return quantizable_layers


class RAWGPTQuantizer(object):
    """Main API for GPTQ algorithm.

    Please refer to:
    GPTQ: Accurate Post-training Compression for Generative Pretrained Transformers
    url: https://arxiv.org/abs/2210.17323
    """

    def __init__(
        self,
        model,
        weight_config={},
        nsamples=128,
        use_max_length=True,
        max_seq_length=2048,
        device=None,
        use_layer_wise=False,
        model_path="",
        quant_lm_head=False,
        dataloader=None,
        use_block_wise=False,
        *args,
        **kwargs,
    ):
        """Init RAWGPTQuantizer.

        Args:
            model: the fp32 model to quantize
            weight_config (dict, optional): contains all info required by GPTQ. Defaults to {}. For example,
            weight_config={
                'layer1':
                {
                    'bits': 4,
                    'group_size': 32,
                    'sym': False,
                    'percdamp': .01,
                    'act_order': False
                }
                ...
            }
            nsamples (int): the number of calibration data samples.
            use_max_length (bool): set all sequence length to be same length.
            max_seq_length (int): the same length of all sequence length.
            dataloader: an iterable containing calibration datasets, contains (inputs, targets)
            use_layer_wise (bool): Whether to load weights and quantize per layer. Defaults to False.
            use_block_wise (bool): Whether to load weights and quantize per block. Defaults to False.
            model_path (str): Model path that is used to load state_dict per layer.
            quant_lm_head (bool): Indicates whether quantize the lm_head layer in transformers. Defaults to False.
            device (str): cpu or cuda.
        """
        # model
        self.model = model
        # self.use_cache = self.model.config.use_cache
        self.gptq_related_blocks = trace_gptq_target_blocks(self.model)  # get the transformer block list above
        self.dtype = next(iter(self.model.parameters())).dtype
        quantizable_layers = log_quantizable_layers_per_transformer(self.gptq_related_blocks)

        # weight config
        self.weight_config = weight_config
        # default settings, check configs
        self.dtype_default = "int"
        self.bits_default = 4
        self.group_size_default = 128
        self.block_size_default = 128
        self.percdamp_default = 0.01
        self.sym_default = False
        self.act_order_default = False
        self.hybrid_order_default = False
        self.fp8_aware_default = False
        self.static_groups_default = False
        self.true_sequential_default = False
        self.quant_lm_head = quant_lm_head
        self.perchannel_default = True
        self.mse_default = False
        self.use_double_quant_default = False
        self.double_quant_dtype_default = "int"
        self.double_quant_bits_default = 4
        self.double_quant_group_size_default = 128
        self.double_quant_sym_default = False
        self.check_layer_config()

        # device
        self.device = get_accelerator(kwargs.pop("device", "auto")).current_device_name()
        self.is_ready = False

        self.use_layer_wise = use_layer_wise
        self.use_block_wise = use_block_wise
        if use_layer_wise:
            self.prepare_layer_wise(model_path, quantizable_layers)

        if self.use_block_wise:
            assert not self.use_layer_wise, "Block-wise and layer-wise quantization cannot be used at the same time."
            self.prepare_layer_wise(model_path, layerwise=False)
        self.model_path = get_path(model_path)

        # dataloader
        self.use_max_length = use_max_length
        self.max_seq_length = max_seq_length
        self.dataloader_original = dataloader
        self.dataloader = []
        self.nsamples = nsamples

    def prepare_layer_wise(self, model_path, indicated_layers=None, layerwise=True):
        """Prepare for layer-wise or blockwise quantization, including registering hooks and setting up the model path.

        Args:
            model_path (str): Model path that is used to load state_dict per layer.
            indicated_layers (list, optional): A list of layer names to apply layer-wise quantization.
                                        If None, all layers will be considered.
                                        Layers not specified in this list will be retained in memory
                                        but will not undergo quantization.
            layerwise (bool): Whether to apply layer-wise quantization.
        """
        import os

        from neural_compressor.torch.algorithms.layer_wise import LWQ_WORKSPACE, get_path, register_weight_hooks

        os.makedirs(LWQ_WORKSPACE, exist_ok=True)
        if model_path == "":
            model_path = self.model.path
        assert model_path, "model_path should not be None."
        self.model_path = get_path(model_path)

        if layerwise:
            register_weight_hooks(
                self.model,
                self.model_path,
                device=self.device,
                clean_weight=True,
                saved_path=LWQ_WORKSPACE,
                indicated_layers=indicated_layers,
            )

    def get_full_layer_name(self, sub_layer_name, block_idx):
        """Get full layer name.

        Args:
            sub_layer_name (str): sub layer name
            block_idx (int): index of block

        Returns:
            str: The full name of layer.
        """
        transformer_name = self.gptq_related_blocks["transformers_name"]
        return ".".join([transformer_name, str(block_idx), sub_layer_name])

    def check_layer_config(self):
        """Copy arguments from weight_config to built-in attributes."""
        for layer_name, config in self.weight_config.items():
            self.weight_config[layer_name]["dtype"] = config.get("dtype", self.dtype_default)
            self.weight_config[layer_name]["bits"] = config.get("bits", self.bits_default)
            self.weight_config[layer_name]["group_size"] = config.get("group_size", self.group_size_default)
            self.weight_config[layer_name]["block_size"] = config.get("block_size", self.group_size_default)
            self.weight_config[layer_name]["percdamp"] = config.get("percdamp", self.percdamp_default)
            self.weight_config[layer_name]["sym"] = config.get("sym", self.sym_default)
            self.weight_config[layer_name]["act_order"] = config.get("act_order", self.act_order_default)
            self.weight_config[layer_name]["hybrid_order"] = config.get("hybrid_order", self.hybrid_order_default)
            self.weight_config[layer_name]["fp8_aware"] = config.get("fp8_aware", self.fp8_aware_default)
            self.weight_config[layer_name]["static_groups"] = config.get("static_groups", self.static_groups_default)
            self.weight_config[layer_name]["true_sequential"] = config.get(
                "true_sequential", self.true_sequential_default
            )
            self.weight_config[layer_name]["perchannel"] = config.get("perchannel", self.perchannel_default)
            self.weight_config[layer_name]["mse"] = config.get("mse", self.mse_default)
            self.weight_config[layer_name]["use_double_quant"] = config.get(
                "use_double_quant", self.use_double_quant_default
            )
            self.weight_config[layer_name]["double_quant_dtype"] = config.get(
                "double_quant_dtype", self.double_quant_dtype_default
            )  # only support int
            self.weight_config[layer_name]["double_quant_bits"] = config.get(
                "double_quant_bits", self.double_quant_bits_default
            )
            self.weight_config[layer_name]["double_quant_group_size"] = config.get(
                "double_quant_group_size", self.double_quant_group_size_default
            )
            self.weight_config[layer_name]["double_quant_sym"] = config.get(
                "double_quant_sym", self.double_quant_sym_default
            )
            if self.weight_config[layer_name]["dtype"] != "int" and "int" in self.weight_config[layer_name]["dtype"]:
                self.weight_config[layer_name]["bits"] = int(self.weight_config[layer_name]["dtype"].lstrip("int"))
                self.weight_config[layer_name]["dtype"] = "int"

    def get_layer_config(self, layer_name):
        """Obtain config for one layer, since GPTQ supports layer-wise config."""
        # First try the exact name matching, if cannot find, use re to search. For example, can support ".*" in op_name
        config = None
        config = self.weight_config.get(layer_name, None)
        if config is not None:
            return config
        else:
            for k, v in self.weight_config.items():
                regex = re.compile(k)
                if len(regex.findall(layer_name)) is not None:
                    config = v
                    return config
                else:
                    pass
        return config

    def track_hidden_states(self, data):
        """Track hidden states.

        Args:
            data (tensor/tuple/list): input data.

        Returns:
            tensor.
        """
        if isinstance(data, torch.Tensor):
            return data
        elif isinstance(data, tuple) or isinstance(data, list):
            return data[0]

    @torch.no_grad()
    def prepare_for_calibration(self):
        """Prepare input calibration data and other attributes which are critical for gptq execution."""
        try:
            self.cache_key_arguments = {
                "batch_num": 0
            }  # a dict of list, keyword arguments ("attention_masks", "position_ids", etc.)
            # Note that the first elements in cache_positional_arguments is main input: hidden_states
            self.cache_positional_arguments = []  # a list of list, positional arguments ("rotary_pos_emb" in chatglm)
            self.is_ready = True
        except:
            logger.warning("GPTQ Quantizer initialization failed!")
            pass

        # critical: hooker function which collects inputs
        def forward(layer, *args, **kwargs):
            # inputs[inputs_info['idx']] = input_ids # TODO solve the problem of batchsize!=1
            self.cache_key_arguments["batch_num"] += 1
            for arg in kwargs:
                # TODO: investigate include parameters
                # each outputs can be different shape, hence also use list to store
                # position_embeddings is a list of two tensors in llama, handle it specially.
                if isinstance(kwargs[arg], torch.Tensor) or arg in ["alibi", "position_embeddings"]:
                    if self.cache_key_arguments.get(arg, None) is None:
                        self.cache_key_arguments[arg] = []
                    self.cache_key_arguments[arg].append(kwargs[arg])
                continue
            # copy positional arguments, positional arguments are sensitive for their order, be cautious!
            # Most models in HF has avoid this, but some models still use positional arguments other than
            # hidden_states, chatglm2-6b etc.
            for idx, item in enumerate(args):
                if (idx + 1) > len(self.cache_positional_arguments):
                    # initialize
                    self.cache_positional_arguments.append([])
                self.cache_positional_arguments[idx].append(item)
            raise ValueError

        # Step1: fetch the embeddings and other layers before the transformer stack.
        if not self.use_layer_wise:  # pragma: no cover
            for embedding_name, embedding_layer in self.gptq_related_blocks["embeddings"].items():
                embedding_layer = embedding_layer.to(self.device)

        # Step2: modify the first transformer block's forward function to obtain inputs for calibration
        if not self.use_layer_wise:  # pragma: no cover
            self.gptq_related_blocks["transformers"][0] = self.gptq_related_blocks["transformers"][0].to(self.device)
        self.forward_cache = self.gptq_related_blocks["transformers"][0].forward
        self.gptq_related_blocks["transformers"][0].forward = partial(
            forward, self.gptq_related_blocks["transformers"][0]
        )
        # Step 3: replace model_forward to avoid ValueError
        self.orig_model_forward_cache = self.model.forward
        model_forward_cache = self.model.forward

        def model_forward(model, *args, **kwargs):
            nonlocal model_forward_cache
            try:
                model_forward_cache(*args, **kwargs)
            except ValueError:
                pass

        self.model.forward = partial(model_forward, self.model)

    @torch.no_grad()
    def remove_prepare_for_calibration(self):
        """Prepare for GPTQ quantization."""
        # output inp data shape
        logger.info("All calibration data's shape =>")
        # check all hidden_states shape
        try:
            for hidden_states in self.cache_positional_arguments[0]:
                logger.info(hidden_states.shape)
        except:
            pass
        logger.info("Done.")

        # Step 4: restore original forward function, relocate layers back to cpu.
        self.model.forward = self.orig_model_forward_cache
        self.gptq_related_blocks["transformers"][0].forward = self.forward_cache
        if not self.use_layer_wise:  # pragma: no cover
            self.gptq_related_blocks["transformers"][0] = self.gptq_related_blocks["transformers"][0].cpu()
            for embedding_name, embedding_layer in self.gptq_related_blocks["embeddings"].items():
                embedding_layer.to(self.device)
        torch.cuda.empty_cache()
        # end
        logger.info("GPTQ quantization prepared.")

    def gather_single_batch_from_dict(self, data_dict, idx):
        """Gather single batch from a dict.

        Args:
            data_dict (dict): data dict.
            idx (int): index

        Returns:
            dict: single batch.
        """
        # obtain a set of keyword input from cache
        single_batch = {}
        for k, v in data_dict.items():
            single_batch[k] = data_dict[k][idx]
        return single_batch

    def gather_single_batch_from_list(self, data_list, idx):
        """Gather single batch from a list.

        Args:
            data_dict (dict): data list.
            idx (int): index

        Returns:
            list: single batch.
        """
        # obtain a set of keyword input from cache
        single_batch = []
        for data_item in data_list:
            single_batch.append(data_item[idx])
        return single_batch

    def update_blockwise_hidden_states(self, outs):
        """Update the blockwise hidden states.

        Args:
            outs: the output of block.
        """
        if "hidden_states" in self.cache_key_arguments:
            self.cache_key_arguments["hidden_states"] = outs[:]
        else:
            self.cache_positional_arguments[0] = outs[:]

    def find_true_sequential_config(self):
        """Find true sequential config.

        Returns:
            bool: True or False.
        """
        for layer_name in self.weight_config:
            if self.weight_config[layer_name].get("true_sequential", None) is not None:
                return self.weight_config[layer_name]["true_sequential"]
        return False

    def analyze_true_sequential(self, module, inputs=None):
        """To obtain the depth of each linear layers in this block.

        Args:
            module (nn.module): block.
            inputs (optional): Defaults to None.

        Returns:
            list: layers grouping into sequentials.
        """
        # to obtain the depth of each linear layers in this block
        # obtain all linear layers' names
        layers = find_layers(module)
        layers = list(layers)
        # group layers into sequentials
        # case 1: query, key and value are calculated from one matrix, bloom, etc..
        if "q" in layers[0].lower() and "k" in layers[0].lower():
            qkv_layers = [layers[0]]
            post_qkv_layers = layers[1:]
        else:
            # case 2: qkv are calculated separately.
            qkv_layers = layers[0:3]
            post_qkv_layers = layers[3:]
        layers.clear()
        layers.append(qkv_layers)
        for layer in post_qkv_layers:
            layers.append([layer])
        return layers

    @torch.no_grad()
    def execute_quantization(self, means=None, stds=None):
        """Run quantization."""
        # Step1: prepare quantization (calibration datasets)

        logger.info("Begin ====>")

        # Step2: run gptq quantization in a transformer block-wise manner.
        gptq_config = {}

        self.true_sequential = self.find_true_sequential_config()
        # automatically get true_sequential
        true_sequential_map = self.analyze_true_sequential(self.gptq_related_blocks["transformers"][0])
        logger.info(f"Sequential Name: {true_sequential_map}")
        tblock_length = len(self.gptq_related_blocks["transformers"])
        for param in self.model.parameters():
            param.requires_grad = False

        cpu_mem_0 = get_used_cpu_mem_MB()

        for block_idx in range(tblock_length):
            start_iter = time.time()
            logger.debug(f"Memory usage increase CPU: {get_used_cpu_mem_MB() - cpu_mem_0}")
            logger.info(f"Quantizing layer {block_idx + 1} / {tblock_length}..")
            transformer_block = self.gptq_related_blocks["transformers"][block_idx]

            # Step2.1: obtain all layers (Linear, Conv2d, etc) in the block which can be quantized.
            # device = 'cpu'

            def find_all_layer_names(module, name=""):
                """Get all layers with target types."""
                if len(list(module.named_children())) == 0:
                    return [name]
                res = []
                for name1, child in module.named_children():
                    res += find_all_layer_names(child, name=name + "." + name1 if name != "" else name1)
                return res

            # block weights are meta tensors, load them from disk
            if self.use_block_wise:
                for n in find_all_layer_names(transformer_block):
                    param_name = f"model.layers.{block_idx}." + n + ".weight"
                    try:
                        value = load_value(self.model, param_name, self.model_path, "cpu")
                        set_module_tensor_to_device(transformer_block.get_submodule(n), "weight", "cpu", value)
                    except:
                        pass  # only load w

            if not self.use_layer_wise:  # pragma: no cover
                # if we do not apply layer-wise feature, we still place the entire block on the GPU
                transformer_block = self.gptq_related_blocks["transformers"][block_idx].to(self.device)
            else:
                transformer_block = self.gptq_related_blocks["transformers"][block_idx]
            # Step2.1: obtain all layers (Linear, Conv2d, etc) in the block which can be quantized.
            sub_layers = find_layers(transformer_block)
            sub_layers_to_quant = {}
            # add true sequential options
            if self.true_sequential is not None and self.true_sequential:
                sequentials = true_sequential_map
            else:
                sequentials = [list(sub_layers.keys())]
            # start to process every layers in a sequential
            for sequential in sequentials:
                logger.info(f"Current quantization sequential: {sequential}")
                sub_layers_to_quant = {}
                sequential_layers = {n: sub_layers[n] for n in sequential}
                for layer_name, layer_obj in sequential_layers.items():
                    # filter sub_layers with included layer_names in self.weight_config
                    full_layer_name = self.get_full_layer_name(layer_name, block_idx)
                    # if self.weight_config.get(full_layer_name, None) == None:
                    if self.get_layer_config(full_layer_name) is None:
                        logger.warning(
                            f"{full_layer_name} can be quantized " + "but excluded from quantization configs."
                        )
                    else:
                        sub_layers_to_quant[layer_name] = layer_obj
                del sequential_layers
                sequential_layers = sub_layers_to_quant
                # Step 2.2: Initialize GPTQ quantizers for collected layers.
                gptq_for_this_block = {}
                # initialize gptq quantizer for every layer in a transformer block
                for layer_name in sequential_layers:
                    # weight_config_this_layer = self.weight_config.get(
                    #     self.get_full_layer_name(layer_name, block_idx), None
                    # )
                    full_layer_name = self.get_full_layer_name(layer_name, block_idx)
                    weight_config_this_layer = self.get_layer_config(full_layer_name)
                    if self.use_layer_wise:  # pragma: no cover
                        W = load_value(self.model, full_layer_name + ".weight", self.model_path, self.device)
                    else:
                        if "hpu" in str(self.device):  # pragma: no cover
                            # [SW-206677] memory is not release when module is moved out of HPU
                            sequential_layers[layer_name] = sequential_layers[layer_name].cpu()
                            W = sequential_layers[layer_name].weight.data.clone()
                            sequential_layers[layer_name] = sequential_layers[layer_name].to("hpu")
                        else:
                            W = sequential_layers[layer_name].weight.data.clone()

                    gptq_for_this_block[layer_name] = GPTQ(sequential_layers[layer_name], W, self.device)
                    # gptq_for_this_block[layer_name].quantizer = Quantizer()
                    gptq_for_this_block[layer_name].quantizer.configure(weight_config_this_layer)

                # Step 2.3: modify forward functions to hook inputs data (used in gptq execution)
                def add_batch(_name):
                    def tmp(_, inp, out):
                        gptq_for_this_block[_name].add_batch(inp[0].data, out.data)  # noqa: F821

                    return tmp

                handles = []  # register handles which add inputs and outputs to gptq object
                for layer_name in sequential_layers:
                    handles.append(sequential_layers[layer_name].register_forward_hook(add_batch(layer_name)))
                batch_num = self.cache_key_arguments.pop("batch_num")
                for j in range(batch_num):
                    cache_keyword_batch = self.gather_single_batch_from_dict(self.cache_key_arguments, j)
                    cache_positional_batch = self.gather_single_batch_from_list(self.cache_positional_arguments, j)
                    out = transformer_block(*cache_positional_batch, **cache_keyword_batch)
                    accelerator.synchronize()
                    out = self.track_hidden_states(out)
                self.cache_key_arguments["batch_num"] = batch_num
                for h in handles:
                    h.remove()
                # Step 2.4: everything is prepared, so start quantization!
                for layer_name in sequential_layers:
                    # weight_config_this_layer = self.weight_config.get(
                    #     self.get_full_layer_name(layer_name, block_idx), None
                    # )
                    weight_config_this_layer = self.get_layer_config(self.get_full_layer_name(layer_name, block_idx))
                    logger.info(f"Quantizing layer {layer_name}")
                    if self.use_layer_wise:  # pragma: no cover
                        full_layer_name = self.get_full_layer_name(layer_name, block_idx)
                        W = load_value(self.model, full_layer_name + ".weight", self.model_path, self.device)
                    else:
                        W = sequential_layers[layer_name].weight.data.clone()
                    accelerator.synchronize()
                    if "hpu" in str(self.device):
                        W = W.to("cpu")
                    scale, scale_bf16_to_fp8, zp, Q = gptq_for_this_block[layer_name].fasterquant(
                        W,
                        blocksize=weight_config_this_layer["block_size"],
                        percdamp=weight_config_this_layer["percdamp"],
                        groupsize=weight_config_this_layer["group_size"],
                        act_order=weight_config_this_layer["act_order"],
                        hybrid_order=weight_config_this_layer["hybrid_order"],
                        fp8_aware=weight_config_this_layer["fp8_aware"],
                        static_groups=weight_config_this_layer["static_groups"],
                    )
                    if self.use_layer_wise:  # pragma: no cover
                        from neural_compressor.torch.algorithms.layer_wise import (
                            LWQ_WORKSPACE,
                            clean_module_weight,
                        )

                        sub_layer = sequential_layers[layer_name]
                        full_layer_name = self.get_full_layer_name(layer_name, block_idx)
                        for n, p in sub_layer.named_parameters():
                            param_name = full_layer_name + "." + n
                            if n == "weight":
                                set_module_tensor_to_device(self.model, param_name, self.device, Q)
                            else:
                                value = load_value(self.model, param_name, self.model_path)
                                set_module_tensor_to_device(self.model, param_name, self.device, value)
                        # sub_layer.weight.data = Q
                        torch.save(sub_layer.state_dict(), LWQ_WORKSPACE + f"/{full_layer_name}.pt")
                        clean_module_weight(sub_layer)
                        del Q
                        gc.collect()
                    else:
                        sequential_layers[layer_name].weight.data = Q
                    gptq_config[self.get_full_layer_name(layer_name, block_idx)] = {"scale": scale}
                    if weight_config_this_layer["fp8_aware"]:
                        gptq_config[self.get_full_layer_name(layer_name, block_idx)][
                            "scale_bf16_to_fp8"
                        ] = scale_bf16_to_fp8
                    if not weight_config_this_layer["sym"]:
                        gptq_config[self.get_full_layer_name(layer_name, block_idx)]["zero"] = zp
                    if weight_config_this_layer["act_order"]:  # save perm for restoring the weights
                        gptq_config[self.get_full_layer_name(layer_name, block_idx)]["perm"] = gptq_for_this_block[
                            layer_name
                        ].perm
                    gptq_for_this_block[layer_name].free()

                # Step 2.5: replace output data with quantized weights
                batch_num = self.cache_key_arguments.pop("batch_num")
                for j in range(batch_num):
                    cache_keyword_batch = self.gather_single_batch_from_dict(self.cache_key_arguments, j)
                    cache_positional_batch = self.gather_single_batch_from_list(self.cache_positional_arguments, j)
                    out = transformer_block(*cache_positional_batch, **cache_keyword_batch)
                    accelerator.synchronize()
                    out = self.track_hidden_states(out)
                    # iteratively replace the input with output, thus layerwise quantization can continue.
                    if "hidden_states" in self.cache_key_arguments:
                        self.cache_key_arguments["hidden_states"][j] = out
                    else:
                        self.cache_positional_arguments[0][j] = out
                self.cache_key_arguments["batch_num"] = batch_num
                if self.use_layer_wise:  # pragma: no cover
                    self.gptq_related_blocks["transformers"][block_idx] = transformer_block
                else:
                    transformer_block = transformer_block.cpu()
                    self.gptq_related_blocks["transformers"][block_idx] = transformer_block
                # Step 2.6: export to compressed model
                for layer_name in sequential_layers:
                    weight_config_this_layer = self.get_layer_config(self.get_full_layer_name(layer_name, block_idx))
                    gptq_scale = gptq_config[self.get_full_layer_name(layer_name, block_idx)]["scale"].cpu()
                    if weight_config_this_layer["fp8_aware"]:
                        gptq_scale_bf16_to_fp8 = gptq_config[self.get_full_layer_name(layer_name, block_idx)][
                            "scale_bf16_to_fp8"
                        ].cpu()
                    else:
                        gptq_scale_bf16_to_fp8 = None
                    if not weight_config_this_layer["sym"]:
                        gptq_zp = gptq_config[self.get_full_layer_name(layer_name, block_idx)]["zero"].cpu()
                    else:
                        gptq_zp = None
                    if weight_config_this_layer["act_order"]:  # save perm for restoring the weights
                        gptq_perm = gptq_config[self.get_full_layer_name(layer_name, block_idx)]["perm"].cpu()
                    else:
                        gptq_perm = None
                    if self.use_layer_wise:
                        state_dict = torch.load(
                            LWQ_WORKSPACE + f"/{self.get_full_layer_name(layer_name, block_idx)}.pt"
                        )
                        Q = state_dict["weight"].data
                        bias = state_dict["bias"] if "bias" in state_dict.keys() else None

                    else:
                        Q = sequential_layers[layer_name].weight.data
                    if weight_config_this_layer["act_order"]:
                        Q.copy_(Q[:, gptq_perm])
                    if is_transformers_imported() and isinstance(sequential_layers[layer_name], transformers.Conv1D):
                        Q = Q.t_().contiguous()
                    from .utility import quant_weight_w_scale

                    Q = quant_weight_w_scale(
                        Q,
                        gptq_scale,
                        gptq_scale_bf16_to_fp8,
                        gptq_zp,
                        weight_config_this_layer["group_size"],
                        dtype=weight_config_this_layer["dtype"],
                        fp8_aware=weight_config_this_layer["fp8_aware"],
                    )
                    if weight_config_this_layer["act_order"]:
                        invperm = torch.argsort(gptq_perm)
                        Q.copy_(Q[:, invperm])
                    int_weight = Q.type(torch.int32)  # copy_ is not workable for different types.
                    # replace module
                    if isinstance(sequential_layers[layer_name], torch.nn.Linear):
                        in_features = sequential_layers[layer_name].in_features
                        out_features = sequential_layers[layer_name].out_features
                    elif is_transformers_imported() and isinstance(sequential_layers[layer_name], transformers.Conv1D):
                        in_features = sequential_layers[layer_name].weight.shape[0]
                        out_features = sequential_layers[layer_name].weight.shape[1]
                        int_weight = sequential_layers[layer_name].weight.t_().contiguous()
                        scale = scale.t_().contiguous()
                        scale_bf16_to_fp8 = scale_bf16_to_fp8.t_().contiguous()
                        zp = zp.t_().contiguous() if zp is not None else zp

                    if not self.use_layer_wise:
                        bias = sequential_layers[layer_name].bias

                    new_module = INCWeightOnlyLinear(
                        in_features,
                        out_features,
                        dtype=weight_config_this_layer["dtype"],
                        bits=weight_config_this_layer["bits"],
                        group_size=weight_config_this_layer["group_size"],
                        zp=gptq_zp is not None,
                        bias=bias is not None,
                        g_idx=gptq_perm is not None,
                        device="cpu",
                    )
                    new_module.pack(
                        int_weight,
                        gptq_scale,
                        gptq_zp,
                        bias,
                        scale_bf16_to_fp8=gptq_scale_bf16_to_fp8,
                        g_idx=gptq_perm,
                    )
                    set_module(transformer_block, layer_name, new_module)
                    accelerator.synchronize()

                if self.use_block_wise:

                    from neural_compressor.torch.algorithms.layer_wise import (
                        LWQ_WORKSPACE,
                        clean_module_weight,
                    )

                    block = self.gptq_related_blocks["transformers"][block_idx]
                    full_block_name = self.gptq_related_blocks["transformers_name"] + "." + str(block_idx)

                    modified_state_dict = {
                        f"{full_block_name}.{key}": value for key, value in block.state_dict().items()
                    }
                    torch.save(modified_state_dict, LWQ_WORKSPACE + f"/{full_block_name}.pt")
                    logger.info(f"Saving block to {LWQ_WORKSPACE + f'/{full_block_name}.pt'}")
                    for n, l in find_all_layers(transformer_block).items():
                        clean_module_weight(l)

                    # in the first iter, save all none-layer weights (embeddings, etc)
                    if block_idx == 0:
                        aux_state_dict = {}
                        state_dict = self.model.state_dict()

                        for key, value in state_dict.items():
                            # Filter out tensors that are on the 'meta' device
                            if value.device.type != "meta":
                                aux_state_dict[key] = value

                        torch.save(aux_state_dict, LWQ_WORKSPACE + "/auxiliaries.pt")

                del gptq_for_this_block
                accelerator.synchronize()
                torch.cuda.empty_cache()
                gc.collect()
                logger.info("------------------------------")

            logger.info(f"Layer quant time = {time.time() - start_iter}")
        # 2.7.1 do the post transformer blocks quantization
        do_post_transformer_quant = self.quant_lm_head
        if do_post_transformer_quant:
            logger.info("Quantizing post transformer layers")
            # the input should be self.cache_key_arguments and self.cache_positional_arguments
            sub_layers = find_layers(self.gptq_related_blocks["transformers_post"]["layer"])
            sub_layers_to_quant = {}
            for layer_name, layer_obj in sub_layers.items():
                # filter sub_layers with included layer_names in self.weight_config
                full_layer_name = self.gptq_related_blocks["transformers_post"]["name"]
                # if self.weight_config.get(full_layer_name, None) == None:
                if self.get_layer_config(full_layer_name) is None:
                    logger.warning(f"{full_layer_name} can be quantized " + "but excluded from quantization configs.")
                else:
                    sub_layers_to_quant[full_layer_name] = layer_obj
            del sub_layers
            sub_layers = sub_layers_to_quant
            gptq_post_block = {}

            def add_batch_post(_name):
                def tmp(_, inp, out):
                    gptq_post_block[_name].add_batch(inp[0].data, out.data)

                return tmp

            for layer_name in sub_layers:
                full_layer_name = self.gptq_related_blocks["transformers_post"]["name"]
                weight_config_this_layer = self.get_layer_config(full_layer_name)
                if self.use_layer_wise:  # pragma: no cover
                    full_layer_name = self.gptq_related_blocks["transformers_post"]["name"]
                    W = load_value(self.model, full_layer_name + ".weight", self.model_path)
                else:
                    if "hpu" in self.device:  # pragma: no cover
                        # [SW-206677] memory is not release when module is moved out of HPU
                        sub_layers[layer_name] = sub_layers[layer_name].cpu()
                        W = sub_layers[layer_name].weight.data.clone()
                        sub_layers[layer_name] = sub_layers[layer_name].to("hpu")
                    else:
                        W = sub_layers[layer_name].weight.data.clone()

                gptq_post_block[layer_name] = GPTQ(sub_layers[layer_name], W, self.device)
                # gptq_for_this_block[layer_name].quantizer = Quantizer()
                gptq_post_block[layer_name].quantizer.configure(weight_config_this_layer)
            # generate the gptq quantizer
            handles = []  # register handles which add inputs and outputs to gptq object
            for layer_name in sub_layers:
                handles.append(sub_layers[layer_name].register_forward_hook(add_batch_post(layer_name)))
            for j in range(len(self.dataloader)):
                if "hidden_states" in self.cache_key_arguments:
                    out = sub_layers[layer_name](self.cache_key_arguments["hidden_states"][j])
                else:
                    out = sub_layers[layer_name](self.cache_positional_arguments[0][j])

            # if "hidden_states" in self.cache_key_arguments:
            #     self.cache_key_arguments["hidden_states"] = outs[:]
            # else:
            #     self.cache_positional_arguments[0] = outs[:]
            # perform the inference process

            for h in handles:
                h.remove()

            for layer_name in sub_layers:
                full_layer_name = self.gptq_related_blocks["transformers_post"]["name"]
                weight_config_this_layer = self.get_layer_config(full_layer_name)
                if "hpu" in self.device:
                    W = W.to("cpu")
                scale, scale_bf16_to_fp8, zp, Q = gptq_post_block[layer_name].fasterquant(
                    W,
                    blocksize=weight_config_this_layer["block_size"],
                    percdamp=weight_config_this_layer["percdamp"],
                    groupsize=weight_config_this_layer["group_size"],
                    act_order=weight_config_this_layer["act_order"],
                    hybrid_order=weight_config_this_layer["hybrid_order"],
                    fp8_aware=weight_config_this_layer["fp8_aware"],
                    static_groups=weight_config_this_layer["static_groups"],
                )
                if self.use_layer_wise:  # pragma: no cover
                    from neural_compressor.torch.algorithms.layer_wise import (
                        LWQ_WORKSPACE,
                        clean_module_weight,
                    )

                    sub_layer = sub_layers[layer_name]
                    full_layer_name = self.gptq_related_blocks["transformers_post"]["name"]
                    for n, p in sub_layer.named_parameters():
                        param_name = full_layer_name + "." + n
                        if n == "weight":
                            set_module_tensor_to_device(self.model, param_name, self.device, Q)
                        else:
                            value = load_value(self.model, param_name, self.model_path, device=self.device)
                            set_module_tensor_to_device(self.model, param_name, self.device, value)
                    # sub_layer.weight.data = Q
                    torch.save(sub_layer.state_dict(), LWQ_WORKSPACE + f"/{full_layer_name}.pt")
                    clean_module_weight(sub_layer)
                    del Q
                    gc.collect()
                else:
                    sub_layers[layer_name].weight.data = Q
                # save the quantization results
                gptq_config[full_layer_name] = {"scale": scale}
                if weight_config_this_layer["fp8_aware"]:
                    gptq_config[full_layer_name]["scale_bf16_to_fp8"] = scale_bf16_to_fp8
                if not weight_config_this_layer["sym"]:
                    gptq_config[full_layer_name]["zero"] = zp
                if weight_config_this_layer["act_order"] and not weight_config_this_layer["static_groups"]:
                    # save perm for restoring the weights, but only when static_groups is not enabled.
                    gptq_config[full_layer_name]["perm"] = gptq_post_block[full_layer_name].perm
                gptq_post_block[layer_name].free()

            # 2.7.2 lm_head: export to compressed model
            for layer_name in sub_layers:
                full_layer_name = self.gptq_related_blocks["transformers_post"]["name"]
                weight_config_this_layer = self.get_layer_config(full_layer_name)
                gptq_scale = gptq_config[full_layer_name]["scale"].cpu()
                if weight_config_this_layer["fp8_aware"]:
                    gptq_scale_bf16_to_fp8 = gptq_config[full_layer_name]["scale_bf16_to_fp8"].cpu()
                else:
                    gptq_scale_bf16_to_fp8 = None
                if not weight_config_this_layer["sym"]:
                    gptq_zp = gptq_config[full_layer_name]["zero"].cpu()
                else:
                    gptq_zp = None
                if weight_config_this_layer["act_order"]:  # save perm for restoring the weights
                    gptq_perm = gptq_config[full_layer_name]["perm"].cpu()
                else:
                    gptq_perm = None
                if self.use_layer_wise:  # pragma: no cover
                    state_dict = torch.load(LWQ_WORKSPACE + f"/{full_layer_name}.pt")
                    Q = state_dict["weight"].data
                    bias = state_dict["bias"] if "bias" in state_dict.keys() else None
                else:
                    Q = sub_layers[layer_name].weight.data
                if weight_config_this_layer["act_order"]:
                    Q.copy_(Q[:, gptq_perm])
                if is_transformers_imported() and isinstance(
                    sub_layers[layer_name], transformers.Conv1D
                ):  # pragma: no cover
                    Q = Q.t_().contiguous()
                from .utility import quant_weight_w_scale

                quant_weight_w_scale(
                    Q,
                    gptq_scale,
                    gptq_scale_bf16_to_fp8,
                    gptq_zp,
                    weight_config_this_layer["group_size"],
                    dtype=weight_config_this_layer["dtype"],
                )
                if weight_config_this_layer["act_order"]:
                    invperm = torch.argsort(gptq_perm)
                    Q.copy_(Q[:, invperm])
                int_weight = Q.type(torch.int32)  # copy_ is not workable for different types.
                # replace module
                if isinstance(sub_layers[layer_name], torch.nn.Linear):
                    in_features = sub_layers[layer_name].in_features
                    out_features = sub_layers[layer_name].out_features
                elif is_transformers_imported() and isinstance(
                    sub_layers[layer_name], transformers.Conv1D
                ):  # pragma: no cover
                    in_features = sub_layers[layer_name].weight.shape[0]
                    out_features = sub_layers[layer_name].weight.shape[1]
                    int_weight = sub_layers[layer_name].weight.t_().contiguous()
                    scale = scale.t_().contiguous()
                    scale_bf16_to_fp8 = scale_bf16_to_fp8.t_().contiguous()
                    zp = zp.t_().contiguous() if zp is not None else zp

                if not self.use_layer_wise:  # pragma: no cover
                    bias = sub_layers[layer_name].bias

                new_module = INCWeightOnlyLinear(
                    in_features,
                    out_features,
                    dtype=weight_config_this_layer["dtype"],
                    bits=weight_config_this_layer["bits"],
                    group_size=weight_config_this_layer["group_size"],
                    zp=gptq_zp is not None,
                    bias=bias is not None,
                    g_idx=gptq_perm is not None,
                    device="cpu",
                )
                new_module.pack(
                    int_weight,
                    gptq_scale,
                    gptq_zp,
                    bias,
                    scale_bf16_to_fp8=gptq_scale_bf16_to_fp8,
                    g_idx=gptq_perm,
                )
                set_module(self.model, layer_name, new_module)

        # Clear temporary workspace
        if self.use_layer_wise:
            import shutil

            shutil.rmtree(LWQ_WORKSPACE, ignore_errors=True)
        logger.info("Quantization done")
        # self.model.config.use_cache = self.use_cache
        return self.model


class GPTQ:
    """Please refer to the following.

    GPTQ: Accurate Post-training Compression for Generative Pretrained Transformers (https://arxiv.org/abs/2210.17323)
    """

    def __init__(self, layer, W, device="cpu"):
        """Init GPTQ."""
        self.layer = layer
        self.device = device
        # W = layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d) or isinstance(self.layer, nn.Conv1d):
            W = W.flatten(1)
        if is_transformers_imported() and isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        self.rows = W.shape[0]  # output channels
        self.columns = W.shape[1]  # input channels
        self.H = torch.zeros((self.columns, self.columns), device=self.device)
        self.nsamples = 0
        self.quantizer = Quantizer()
        self.perm = None  # act_order choice

    def add_batch(self, inp, out):
        """Add inputs and outputs to gptq object."""
        # if DEBUG:
        #     self.inp1 = inp
        #     self.out1 = out
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.layer, nn.Linear) or (
            is_transformers_imported() and isinstance(self.layer, transformers.Conv1D)
        ):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()
        # TODO: llm's transformer sequential with nn.conv2d is currently not under test
        # if isinstance(self.layer, nn.Conv2d):
        #     unfold = nn.Unfold(
        #         self.layer.kernel_size,
        #         dilation=self.layer.dilation,
        #         padding=self.layer.padding,
        #         stride=self.layer.stride
        #     )
        #     inp = unfold(inp)
        #     inp = inp.permute([1, 0, 2])
        #     inp = inp.flatten(1)
        self.H *= self.nsamples / (self.nsamples + tmp)
        self.nsamples += tmp
        # inp = inp.float()
        inp = math.sqrt(2 / self.nsamples) * inp.float()
        # self.H += 2 / self.nsamples * inp.matmul(inp.t())
        self.H += inp.matmul(inp.t())  # H = X*X, which should be a sym matrix

    def fasterquant(
        self,
        W,
        blocksize=128,
        percdamp=0.01,
        groupsize=-1,
        act_order=False,
        hybrid_order=False,
        fp8_aware=False,
        static_groups=False,
    ):
        """Run quantization.

        Args:
            W (tensor): weight tensor.
            block_size (int): Execute quantization per block, block shape = [C_out, block_size]. Default to 128.
            percdamp (float): percdamp (float): Percentage of Hessian's diagonal values' average, which will be added
                                to Hessian's diagonal to increase numerical stability. Defaults to 0.01.
            groupsize (int): Size of weight groups. Defaults to -1.
            act_order (bool): Whether to sort Hessian's diagonal values to rearrange channel-wise quantization order.
                                Defaults to False.
            static_groups (bool): Whether to calculate group wise quantization parameters in advance. This option
                                    mitigate actorder's extra computational requirements. Default to False.

        Returns:
            scale, scale_bf16_to_fp8, zero, Q
        """
        # W = self.layer.weight.data.clone()
        weight_shape, weight_dtype = W.shape, W.data.dtype
        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        if is_transformers_imported() and isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        W = W.float()

        tick = time.time()

        if not self.quantizer.ready():
            self.quantizer.find_params(W, weight=True)

        H = self.H
        if "hpu" in str(self.device):
            H = H.to("cpu")
        del self.H
        dead = torch.diag(H) == 0
        H[dead, dead] = 1
        W[:, dead] = 0  # such channel makes no contribution to quantization computation

        # enable static_groups
        # calculate the quantization parameters for original group in advance.
        if static_groups:
            import copy

            groups = []
            for i in range(0, self.columns, groupsize):
                quantizer = copy.deepcopy(self.quantizer)
                quantizer.find_params(W[:, i : (i + groupsize)], weight=True)
                groups.append(quantizer)

        # rearrange considering diag's value without mixing groups (for faster inference)
        if hybrid_order:
            assert not act_order, "Error: hybrid_act_order is not allowed with act_order"
            local_perms = self.quantizer.compute_local_perms(torch.diag(H), groupsize)
            global_perm = self.quantizer.compute_global_perm(torch.diag(H), groupsize)
            final_perm = self.quantizer.compose_final_perm(local_perms, global_perm, groupsize)
            W = W[:, final_perm]
            H = H[final_perm][:, final_perm]

        # rearrange considering the diag's value
        if act_order:
            perm = torch.argsort(torch.diag(H), descending=True)
            W = W[:, perm]
            H = H[perm][:, perm]
            self.perm = perm.clone()

        Losses = torch.zeros_like(W)
        Q = torch.zeros_like(W)

        damp = percdamp * torch.mean(torch.diag(H))
        # TODO: [SW-201115] when index device is not the same as tensor, the H[diag, diag] += damp doesn't effect.
        if "hpu" in str(self.device):
            diag = torch.arange(self.columns, device="cpu")
        else:
            diag = torch.arange(self.columns, device=self.device)
        H[diag, diag] += damp  # add a average value of
        H = torch.linalg.cholesky(H)
        H = torch.cholesky_inverse(H)
        H = torch.linalg.cholesky(H, upper=True)
        Hinv = H

        scale = []
        zero = []
        scale_bf16_to_fp8 = torch.tensor([-1])  # initialization

        if fp8_aware:
            fullscale = torch.finfo(
                torch.float8_e4m3fnuz
            ).max  # Use Gaudi2's dynamic range as it is lower than Gaudi3's
            # protective range is added to prevent dequant overflow from INT4 to FP8
            protective_range = fullscale - (fullscale / self.quantizer.maxq)
            self.quantizer.find_params_fp8(
                W, fullscale=protective_range, scaling_method="pow2"
            )  # scaling method is hard coded
            scale_bf16_to_fp8 = torch.atleast_1d(self.quantizer.scale_bf16_to_fp8)
            W_8 = torch.clamp(W / scale_bf16_to_fp8, min=-protective_range, max=protective_range)
            W_8 = W_8.to(torch.float8_e4m3fn).float()

        for i1 in range(0, self.columns, blocksize):
            i2 = min(i1 + blocksize, self.columns)
            count = i2 - i1

            W1 = W[:, i1:i2].clone()
            Q1 = torch.zeros_like(W1)
            Err1 = torch.zeros_like(W1)
            Losses1 = torch.zeros_like(W1)
            Hinv1 = Hinv[i1:i2, i1:i2]

            for i in range(count):  # within a block, channel wise
                w = W1[:, i]
                d = Hinv1[i, i]

                if groupsize != -1:
                    if not static_groups:
                        if (i1 + i) % groupsize == 0:
                            if fp8_aware:
                                self.quantizer.find_params(W_8[:, (i1 + i) : (i1 + i + groupsize)], weight=True)
                            else:
                                self.quantizer.find_params(W[:, (i1 + i) : (i1 + i + groupsize)], weight=True)
                            scale.append(self.quantizer.scale)
                            zero.append(self.quantizer.zero)
                    else:
                        idx = i1 + i
                        if act_order:
                            idx = perm[idx]
                        self.quantizer = groups[idx // groupsize]
                if fp8_aware:
                    w_8 = w / scale_bf16_to_fp8
                    hat_w_8 = self.quantizer.quantize(
                        w_8.unsqueeze(1), self.quantizer.scale, self.quantizer.zero, self.quantizer.maxq
                    )
                    hat_w_8 = (
                        torch.clamp(hat_w_8, min=-protective_range, max=protective_range)
                        .to(torch.float8_e4m3fn)
                        .float()
                        .flatten()
                    )
                    q = scale_bf16_to_fp8 * hat_w_8
                else:
                    q = self.quantizer.quantize(
                        w.unsqueeze(1), self.quantizer.scale, self.quantizer.zero, self.quantizer.maxq
                    ).flatten()
                Q1[:, i] = q
                Losses1[:, i] = (w - q) ** 2 / d**2

                err1 = (w - q) / d
                W1[:, i:] -= err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
                Err1[:, i] = err1

            Q[:, i1:i2] = Q1
            Losses[:, i1:i2] = Losses1 / 2

            W[:, i2:] -= Err1.matmul(Hinv[i1:i2, i2:])
            # if fp8_aware:
            #     W_8 = torch.clamp(W / scale_bf16_to_fp8, min=-protective_range, max=protective_range)
            #     W_8 = W_8.to(torch.float8_e4m3fn).float()

            # if DEBUG:
            #     self.layer.weight.data[:, :i2] = Q[:, :i2]
            #     self.layer.weight.data[:, i2:] = W[:, i2:]
            #     logger.info(f"{torch.sum((self.layer(self.inp1) - self.out1) ** 2)}")
            #     logger.info(f"{torch.sum(Losses)}")

        if str(self.device).startswith("cuda"):
            torch.cuda.synchronize()
        logger.info(f"time {(time.time() - tick)}")
        logger.info(f"error {torch.sum(Losses).item()}")

        if hybrid_order:
            inv_final = self.quantizer.invert_perm(final_perm)
            Q = Q[:, inv_final]
            inv_global_perm = self.quantizer.invert_perm(global_perm)
            inv_global_perm_list = inv_global_perm.tolist()
            temp_scale = [scale[i] for i in inv_global_perm_list]
            scale = temp_scale
            temp_zero = [zero[i] for i in inv_global_perm_list]
            zero = temp_zero

        if act_order:
            invperm = torch.argsort(perm)
            Q = Q[:, invperm]

        if is_transformers_imported() and isinstance(self.layer, transformers.Conv1D):
            Q = Q.t()
        # self.layer.weight.data = Q.reshape(self.layer.weight.shape).to(self.layer.weight.data.dtype)
        Q = Q.reshape(weight_shape).to(weight_dtype)
        if DEBUG:
            logger.info(f"{torch.sum((self.layer(self.inp1) - self.out1) ** 2)}")

        if scale == []:
            scale.append(self.quantizer.scale)
            zero.append(self.quantizer.zero)
        scale = torch.cat(scale, dim=1)
        zero = torch.cat(zero, dim=1)
        if "hpu" in str(self.device):  # pragma: no cover
            scale = scale.to(self.device)
            scale_bf16_to_fp8 = scale_bf16_to_fp8.to(self.device)
            zero = zero.to(self.device)
            Q = Q.to(self.device)
        return scale, scale_bf16_to_fp8, zero, Q

    def free(self):
        """Free memory."""
        if DEBUG:
            self.inp1 = None
            self.out1 = None
        self.H = None
        self.Losses = None
        self.Trace = None
        torch.cuda.empty_cache()


class Quantizer(nn.Module):
    """Quantizer."""

    def __init__(self, shape=1):
        """Init Quantizer."""
        super(Quantizer, self).__init__()
        self.maxq = 0
        self.register_buffer("scale", torch.zeros(shape))
        self.register_buffer("zero", torch.zeros(shape))
        self.register_buffer("scale_bf16_to_fp8", torch.zeros(shape))

    def configure(self, weight_config_this_layer, norm=2.4, grid=100, maxshrink=0.8, trits=False):
        """Configure the quantizer."""
        for k, v in weight_config_this_layer.items():
            setattr(self, k, v)
        # self.maxq = torch.tensor(2**self.bits - 1)
        self.maxq = 2**self.bits - 1
        self.scheme = "sym" if self.sym else "asym"
        self.double_quant_scheme = "sym" if self.double_quant_sym else "asym"
        self.norm = norm
        self.grid = grid
        self.maxshrink = maxshrink
        if trits:
            self.maxq = -1

    def compute_local_perms(self, diag_H, groupsize):
        """Compute local permutations for each group of weights.

        For each group, compute a permutation that orders the indices in descending order based on
        the corresponding diagonal values of H.

        Args:
            diag_H (Tensor): 1D tensor representing the diagonal of the Hessian.
            groupsize (int): Number of columns/weights per group.

        Returns:
            local_perms (list of Tensors): Each element is a permutation (indices) for that group.
        """
        n = diag_H.numel()
        num_groups = n // groupsize
        local_perms = []
        for g in range(num_groups):
            start = g * groupsize
            end = start + groupsize
            sub_diag = diag_H[start:end]
            # Get local permutation: indices that would sort sub_diag in descending order.
            local_perm = torch.argsort(sub_diag, descending=True)
            local_perms.append(local_perm)
        return local_perms

    def compute_global_perm(self, diag_H, groupsize):
        """Compute a permutation for the groups themselves.

        Here we choose the maximum diagonal value within each group as the group metric
        and sort the groups in descending order.

        Args:
            diag_H (Tensor): 1D tensor representing the diagonal of the Hessian.
            groupsize (int): Number of columns/weights per group.

        Returns:
            global_perm (Tensor): 1D tensor of length num_groups with the new order of groups.
        """
        n = diag_H.numel()
        num_groups = n // groupsize
        group_metric = []
        for g in range(num_groups):
            start = g * groupsize
            end = start + groupsize
            group_metric.append(diag_H[start:end].max().item())
        # Create a tensor on the same device as diag_H.
        group_metric = torch.tensor(group_metric, device=diag_H.device)
        global_perm = torch.argsort(group_metric, descending=True)
        return global_perm

    def compose_final_perm(self, local_perms, global_perm, groupsize):
        """Compose the final overall permutation from the local and global permutations.

        Args:
            local_perms (list of Tensors): Local permutation for each group.
            global_perm (Tensor): Global group permutation.
            groupsize (int): Number of indices per group.

        Returns:
            final_perm (Tensor): 1D tensor that maps original indices to new positions.
        """
        num_groups = len(local_perms)
        final_perm = []
        # Process groups in the order specified by global_perm.
        for new_group in range(num_groups):
            # Get the original group index.
            orig_group = global_perm[new_group].item()
            offset = orig_group * groupsize
            local_perm = local_perms[orig_group]
            # Adjust local indices to the full index space.
            for idx in local_perm:
                final_perm.append(idx.item() + offset)
        return torch.tensor(final_perm, dtype=torch.long)

    def invert_perm(self, perm):
        """Compute the inverse of a permutation vector.

        Args:
            perm (Tensor): A 1D tensor containing a permutation of indices.

        Returns:
            inv (Tensor): The inverse permutation such that inv[perm] == torch.arange(len(perm)).
        """
        inv = torch.empty_like(perm)
        inv[perm] = torch.arange(perm.numel(), device=perm.device)
        return inv

    def find_params_fp8(self, x, fullscale=torch.tensor(torch.finfo(torch.float8_e4m3fnuz).max), scaling_method="pow2"):
        """Find scales for bf16 to fp8 quantization."""
        x_maxabs = torch.max(torch.abs(x))
        scale = x_maxabs / (fullscale)
        min_scaling_factor = float(1.0 / (fullscale * 512.0))  ##hard coded, copy from vllm, also appears in AutoRound
        scale = torch.clip(scale, min=min_scaling_factor)
        if scaling_method == "arbitrary":
            self.scale_bf16_to_fp8 = scale.to(torch.bfloat16).float()
            return
        scale = 2 ** torch.ceil(torch.log2(scale))
        if scaling_method == "pow2":
            self.scale_bf16_to_fp8 = scale.to(torch.bfloat16).float()
            return
        if scaling_method == "hw":
            min_scale, max_scale, scale_factor = 0.00390625, 16, 4
            scale = torch.minimum(
                torch.maximum(
                    2 ** (torch.ceil(torch.log2(scale) / scale_factor) * scale_factor),
                    torch.tensor(min_scale, dtype=scale.dtype, device=scale.device),
                ),
                torch.tensor(max_scale, dtype=scale.dtype, device=scale.device),
            )
            self.scale_bf16_to_fp8 = scale.to(torch.bfloat16).float()
            return

    def find_params(self, x, weight=False):
        """Find scale and zero for weight."""
        dev = x.device
        # NF4 FP4
        if self.dtype != "int":
            from .utility import quant_tensor

            tmp = x.clone()  # tmp will be replaced after quant_tensor
            _, scale, zero = quant_tensor(
                tmp,
                dtype=self.dtype,
                bits=self.bits,
                group_size=self.group_size,
                scheme=self.scheme,
                quantile=1.0,
                return_int=True,
                full_range=False,
                double_quant=self.use_double_quant,
                double_quant_dtype=self.double_quant_dtype,
                double_quant_bits=self.double_quant_bits,
                double_quant_scheme=self.double_quant_scheme,
                double_quant_group_size=self.double_quant_group_size,
                double_quant_return_int=False,
            )
            self.scale = scale
            self.zero = torch.zeros_like(scale)
            return
        # INT
        shape = x.shape
        if self.perchannel:
            if weight:
                x = x.flatten(1)
            else:
                if len(shape) == 4:
                    x = x.permute([1, 0, 2, 3])
                    x = x.flatten(1)
                if len(shape) == 3:
                    x = x.reshape((-1, shape[-1])).t()
                if len(shape) == 2:
                    x = x.t()
        else:
            x = x.flatten().unsqueeze(0)

        tmp = torch.zeros(x.shape[0], device=dev)
        xmin = torch.minimum(x.min(1)[0], tmp)
        xmax = torch.maximum(x.max(1)[0], tmp)

        if self.sym:
            xmax = torch.maximum(torch.abs(xmin), xmax)
            tmp = xmin < 0
            if torch.any(tmp):
                xmin[tmp] = -xmax[tmp]
        tmp = (xmin == 0) & (xmax == 0)
        xmin[tmp] = -1
        xmax[tmp] = +1

        if self.maxq < 0:
            self.scale = xmax
            self.zero = xmin
        else:
            self.scale = (xmax - xmin) / self.maxq
            if self.sym:
                self.zero = torch.full_like(self.scale, (self.maxq + 1) / 2)
            else:
                self.zero = torch.round(-xmin / self.scale)

        if self.mse:
            best = torch.full([x.shape[0]], float("inf"), device=dev)
            for i in range(int(self.maxshrink * self.grid)):
                p = 1 - i / self.grid
                xmin1 = p * xmin
                xmax1 = p * xmax
                scale1 = (xmax1 - xmin1) / self.maxq
                zero1 = torch.round(-xmin1 / scale1) if not self.sym else self.zero
                q = self.quantize(x, scale1.unsqueeze(1), zero1.unsqueeze(1), self.maxq)
                q -= x
                q.abs_()
                q.pow_(self.norm)
                err = torch.sum(q, 1)
                tmp = err < best
                if torch.any(tmp):
                    best[tmp] = err[tmp]
                    self.scale[tmp] = scale1[tmp]
                    self.zero[tmp] = zero1[tmp]
        if not self.perchannel:
            if weight:
                tmp = shape[0]
            else:
                tmp = shape[1] if len(shape) != 3 else shape[2]
            self.scale = self.scale.repeat(tmp)
            self.zero = self.zero.repeat(tmp)

        if weight:
            shape = [-1] + [1] * (len(shape) - 1)
            self.scale = self.scale.reshape(shape)
            self.zero = self.zero.reshape(shape)

            if self.use_double_quant:
                # for INT
                from .utility import quant_tensor

                orig_scale_shape = self.scale.shape
                self.scale = self.scale.reshape(1, -1)
                quant_tensor(
                    self.scale,
                    dtype=self.double_quant_dtype,
                    bits=self.double_quant_bits,
                    group_size=self.double_quant_group_size,
                    scheme=self.double_quant_scheme,
                    quantile=1.0,
                    return_int=False,
                    full_range=False,
                )
                self.scale = self.scale.reshape(orig_scale_shape)
            return
        if len(shape) == 4:
            self.scale = self.scale.reshape((1, -1, 1, 1))
            self.zero = self.zero.reshape((1, -1, 1, 1))
        if len(shape) == 3:
            self.scale = self.scale.reshape((1, 1, -1))
            self.zero = self.zero.reshape((1, 1, -1))
        if len(shape) == 2:
            self.scale = self.scale.unsqueeze(0)
            self.zero = self.zero.unsqueeze(0)

    def quantize(self, x, scale, zero, maxq):
        """Do quantization."""
        if self.dtype != "int":
            from .utility import quantize_4bit

            tmp = x.clone()  # tmp will be replaced after quant_tensor
            return quantize_4bit(tmp, dtype=self.dtype, scale=scale)
        else:
            if maxq < 0:
                return (x > scale / 2).float() * scale + (x < zero / 2).float() * zero
            q = torch.clamp(torch.round(x / scale) + zero, 0, maxq)
            return scale * (q - zero)

    def ready(self):
        """Quantizer is ready.

        Returns:
            bool: True or False.
        """
        return torch.all(self.scale != 0)


from neural_compressor.torch.algorithms import Quantizer as INCQuantizer


class GPTQuantizer(INCQuantizer):
    """GPTQ Quantizer."""

    def __init__(self, quant_config={}):
        """Init a GPTQQuantizer object.

        Args:
            quant_config (OrderedDict, optional): quantization config for ops. Defaults to {}.
        """
        super().__init__(quant_config)

    @torch.no_grad()
    def prepare(
        self,
        model,
        nsamples=128,
        max_seq_length=2048,
        use_max_length=True,
        device=None,
        use_layer_wise=False,
        model_path=None,
        quant_lm_head=False,
        use_block_wise=False,
        *args,
        **kwargs,
    ):
        """Run weight-only quantization with."""
        # TODO: unify weight_config keys, add docstring, and support default config
        assert isinstance(model, torch.nn.Module), "only support torch module"
        if use_layer_wise:  # pragma: no cover
            assert model_path is not None, "model_path should not be None when use layer wise mode"

        self.model_device = get_model_device(model)  # return model on the same device
        self.gptq_quantizer = RAWGPTQuantizer(
            model,
            weight_config=self.quant_config,
            nsamples=nsamples,
            use_max_length=use_max_length,
            max_seq_length=max_seq_length,
            device=device,
            use_layer_wise=use_layer_wise,
            model_path=model_path,
            quant_lm_head=quant_lm_head,
            use_block_wise=use_block_wise,
        )
        self.gptq_quantizer.prepare_for_calibration()
        return self.gptq_quantizer.model

    @torch.no_grad()
    def convert(self, model, *args, **kwargs):
        """Convert the prepared model to a quantized model.

        Args:
            model (torch.nn.Module): the prepared model

        Returns:
            The quantized model.
        """
        self.gptq_quantizer.model = model
        self.gptq_quantizer.remove_prepare_for_calibration()

        q_model = self.gptq_quantizer.execute_quantization()
        if not self.gptq_quantizer.use_layer_wise and not self.gptq_quantizer.use_block_wise:
            q_model = q_model.to(self.model_device)
        logger.info("GPTQ quantizing done.")
        return q_model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import os
import sys

import numpy as np


def fix_cache_inputs(json_data):
    for layer_index in range(len(json_data["Nodes"])):
        matmul_av_input = None
        v_cache_input = None
        matmul_qk_input = None
        k_cache_input = None

        for node_name, node_info in json_data["Nodes"].items():
            if f"model.layers.{layer_index}.self_attn.attn.impl.matmul_av" in node_name:
                matmul_av_input = node_info["inputs"][1]
            if f"model.layers.{layer_index}.self_attn.attn.impl.v_cache" in node_name:
                v_cache_input = node_info["inputs"][0]
            if f"model.layers.{layer_index}.self_attn.attn.impl.matmul_qk" in node_name:
                matmul_qk_input = node_info["inputs"][1]
            if f"model.layers.{layer_index}.self_attn.attn.impl.k_cache" in node_name:
                k_cache_input = node_info["inputs"][0]
        if matmul_av_input != v_cache_input:
            json_data["Nodes"][f"model.layers.{layer_index}.self_attn.attn.impl.matmul_av"]["inputs"][1] = v_cache_input
        if matmul_qk_input != k_cache_input:
            json_data["Nodes"][f"model.layers.{layer_index}.self_attn.attn.impl.matmul_qk"]["inputs"][1] = k_cache_input

    return json_data


def parse_args(args):
    parser = argparse.ArgumentParser(
        description="Run the measurements parser",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-m",
        "--measurements",
        type=str,
        help="full path to the directory of the measurements that should be fixed",
    )
    parser.add_argument(
        "-o",
        "--out",
        type=str,
        default="",
        help="path to the directory where the fixed measurements will be written. "
             "if not provided will be set to the measurements directory.",
    )
    parser.add_argument(
        "-k",
        "--keep-orig",
        action="store_true",
        default=False,
        help="whether to keep or remove original measurement files",
    )
    return parser.parse_args(args)


def main(args):
    args = parse_args(args)
    measurements_path = args.measurements
    output_path = args.out
    same_dir = False
    if output_path == "":
        output_path = measurements_path
        same_dir = True
    elif not os.path.exists(output_path):
        os.makedirs(output_path)
    # If output path is provided we assume it is different than the measurements path.
    remove_orig = not args.keep_orig
    measurements_paths = os.listdir(measurements_path)
    measurements_paths_ranges = [
        measurement_path
        for measurement_path in measurements_paths
        if measurement_path.endswith(".json")
        and "MAXABS_HW" not in measurement_path
        and "mod_list" not in measurement_path
    ]
    measurements_paths_scales = [
        measurement_path
        for measurement_path in measurements_paths
        if measurement_path.endswith(".json")
        and "MAXABS_HW" in measurement_path
        and "mod_list" not in measurement_path
    ]

    for measurement in measurements_paths_ranges + measurements_paths_scales:
        fixed_json_path = os.path.join(output_path, f"{measurement.split(os.sep)[-1]}")
        path = os.path.join(measurements_path, measurement)
        npz_path = path.replace(".json", ".npz")
        if same_dir:
            renamed_path = os.path.join(measurements_path, f"orig_{measurement}")
            renamed_npz_path = renamed_path.replace(".json", ".npz")
            os.rename(path, renamed_path)
            os.rename(npz_path, renamed_npz_path)
            path = renamed_path
            npz_path = renamed_npz_path
        with open(fixed_json_path, "w") as fixed_json_file:
            with open(path, "r") as json_file:
                data_to_fix = json.load(json_file)
                fixed_data = fix_cache_inputs(data_to_fix)

                json.dump(fixed_data, fixed_json_file)

                global_rank = fixed_data["GlobalRank"]
                local_rank = fixed_data["LocalRank"]
                mode = fixed_data["Mode"]
                nodes = fixed_data["Nodes"]
                layers = {}
                fixed_npz_path = fixed_json_path.replace(".json", ".npz")
                for layer, dlayer in nodes.items():
                    layers[layer] = {}
                    layers[layer]["inputs"] = [np.array(x) for x in dlayer["inputs"]]
                    if dlayer.get("outputs") is not None:
                        layers[layer]["outputs"] = [np.array(x) for x in dlayer["outputs"]]
                    if dlayer.get("params") is not None and dlayer["params"].get("weight") is not None:
                        layers[layer]["params"] = {}
                        layers[layer]["params"]["weight"] = np.array(dlayer["params"]["weight"])
                df = {"GlobalRank": global_rank, "LocalRank": local_rank, "Mode": mode, "Nodes": layers}
                with open(fixed_npz_path, "w"):
                    np.savez(fixed_npz_path, df)
        if remove_orig:
            os.remove(path)
            os.remove(npz_path)

    print("Finished postprocessing_measurements script")


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
from threading import Thread

from neural_compressor.torch.algorithms.fp8_quant.utils.logger import logger
from neural_compressor.torch.algorithms.fp8_quant._quant_common.quant_config import get_hqt_config
from .save_files import (
    create_files_names, measure_control_to_state_dict, save_measurements_files, save_measurements,
    get_mod_extra_config_dict, gmod_list )


def dump_direct_call(model):
    save_measurements(model)

def dump_threading(model):
    t = Thread(target=save_measurements, args=(model,), daemon=True)
    t.start()

def dump_async_call(model):
    asyncio.run(dump_async_call_inner(model))

async def dump_async_call_inner(model):
    mcd = get_mod_extra_config_dict(model)
    await save_measurements_async_wrapper(model, mcd)

async def save_measurements_async_wrapper(model, mcd):
    config = get_hqt_config(model).cfg
    fname_base, fname_np, fname_list, measure_type = create_files_names(config)
    sd, sdl = measure_control_to_state_dict(mcd)
    save_measurements_files(model, sd, sdl, gmod_list, fname_np, fname_list, fname_base, measure_type)


def dump_shelv(model):
    pass

_measurement_dump_method = os.getenv("MEASUREMENT_DUMP_METHOD", "1")
_measurement_dump_method_dict = {
    "1": dump_direct_call,
    # below methods shouldn't be currently used as they are not fully completed
    "2": dump_threading,
    "3": dump_async_call,
    "5": dump_shelv
}


def _increment_calibration_samples_counter(model, *args):  # post hook function
    model.calibration_samples_counter += 1
    if model.calibration_samples_counter % model.calibration_sample_interval == 0:
        logger.debug("Reached sampling interval limit: %d, total samples: %d, dumping measurements.",
                     model.calibration_sample_interval, model.calibration_samples_counter)
        _measurement_dump_method_dict[_measurement_dump_method](model)
        logger.debug("finished dumping measurements.")

def add_calibration_samples_counter(model_to_calibrate, calibration_sample_interval):
    """
    Adds a forward post-hook to the model that counts the number of calibration samples processed.
    When the maximum number of samples is reached, it saves the measurements.
    """
    model_to_calibrate.calibration_samples_counter = 0
    model_to_calibrate.calibration_sample_interval = calibration_sample_interval
    model_to_calibrate.register_forward_hook(_increment_calibration_samples_counter)
    logger.info("Calibration samples interval added to the model - %d.", calibration_sample_interval)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This utility is for block-wise calibration of LLMs."""

import gc
from functools import partial

import torch

from neural_compressor.torch.utils import (
    fetch_module,
    forward_wrapper,
    get_accelerator,
    get_non_persistent_buffers,
    load_non_persistent_buffers,
    logger,
    set_module,
)

cur_accelerator = get_accelerator()


# copied from neural_compressor.torch.algorithms.weight_only.utility
def get_block_prefix(model):
    """Get prefix and number of attention blocks of transformer models.

    Args:
        model (torch.nn.Module): input model

    Returns:
        block_prefix(str): block_list name in model
        block_num(int): number of block in block_list
    """
    module_types = [torch.nn.ModuleList]
    for n, m in model.named_modules():
        if type(m) in module_types:
            block_prefix = n
            block_num = len(m)
            logger.debug(f"block_prefix: {block_prefix}, block_num: {block_num} ")
            break
    assert block_num > 0, "block num shouldn't be zero!"
    return block_prefix, block_num


def replace_forward(model):
    """Replace forward to get the input args and kwargs of first block for AWQ algorithm.

    Args:
        model (torch.nn.Module): input model.

    Raises:
        ValueError: to avoid inference of rest parts in model.

    Returns:
        torch.nn.Module: model with replaced forward.
    """
    # Step 1: replace block_forward to collect block inputs and avoid entire inference
    setattr(model, "total_block_args", [])
    setattr(model, "total_block_kwargs", [])

    def forward(layer, *args, **kwargs):
        # update total_hidden_states, total_block_kwargs, per batch
        model.total_block_args.append(list(args))
        model.total_block_kwargs.append(kwargs)
        raise ValueError

    block_prefix, block_num = get_block_prefix(model)
    block_list = fetch_module(model, block_prefix)
    first_block = block_list[0]
    first_block.forward_orig = first_block.forward
    first_block.forward = partial(forward, first_block)

    # Step 2: replace model_forward to avoid ValueError
    model.forward_orig = model.forward
    model_forward_cache = model.forward

    def model_forward(model, *args, **kwargs):
        nonlocal model_forward_cache
        try:
            model_forward_cache(*args, **kwargs)
        except ValueError:
            pass

    model.forward = partial(model_forward, model)
    return model


def recover_forward(model):
    """Recover model and block forward for AWQ algorithm.

    Args:
        model (torch.nn.Module): input model.

    Returns:
        torch.nn.Module: model with recovered forward.
    """
    model.forward = model.forward_orig

    block_prefix, _ = get_block_prefix(model)
    block_list = fetch_module(model, block_prefix)
    first_block = block_list[0]
    first_block.forward = first_block.forward_orig
    return model


def block_wise_calibration(model, dataloader=None, data=None, inference_dtype=torch.bfloat16):
    """Calibration model on hpu block-by-block to reduce device memory usage.

    Args:
        model (torch.nn.Module): prepared model.
        dataloader (obj): dataloader.
        data (obj): one data.
    """

    def get_block_kwargs(model, dataloader=None, data=None, device="cpu"):
        """Get the input args and kwargs of the first block.

        Args:
            model (torch.nn.Module): prepared model.
            dataloader (obj): dataloader.
            data (obj): one data.
            device (str, optional): target device. Defaults to "cpu".

        Returns:
            total_block_args: list of input args of each batch.
            total_block_kwargs: list of input kwargs of each batch.
        """
        with torch.no_grad():
            model = replace_forward(model)
            mapped_state_dict = model.state_dict()  # using memory mapping
            for n, m in model.named_modules():
                if isinstance(m, torch.nn.Embedding):
                    m = m.to(device).to(inference_dtype)
                    set_module(model, n, m)
            if dataloader is not None:
                for i, data in enumerate(dataloader):
                    logger.info(f"Collecting block inputs from batch: {i}")
                    forward_wrapper(model, data)
                    cur_accelerator.synchronize()
            if data is not None:
                forward_wrapper(model, data)
            model = recover_forward(model)
            for n, m in model.named_modules():
                if isinstance(m, torch.nn.Embedding):
                    m = m.to("meta")  # clean device memory
                    set_module(model, n, m)
            gc.collect()
            cur_accelerator.synchronize()
            total_block_args = getattr(model, "total_block_args", [])
            total_block_kwargs = getattr(model, "total_block_kwargs", [])
            model.load_state_dict(mapped_state_dict, assign=True)  # recover memory mapping
        return total_block_args, total_block_kwargs

    def calibrate_block_and_update_inputs(block, total_block_args, total_block_kwargs, device="cpu"):
        """Calibrate current block and update args and kwargs for the next block.

        Args:
            raw_block (torch.nn.Module): block module for calibration
            total_block_args: list of input args of each batch.
            total_block_kwargs: list of input kwargs of each batch.
            device (str, optional): target device. Defaults to "cpu".

        Returns:
            total_block_args: list of input args of each batch for the next block.
            total_block_kwargs: list of input kwargs of each batch for the next block.
        """
        with torch.no_grad():
            mapped_state_dict = block.state_dict()  # using memory mapping
            block = block.to(device).to(inference_dtype)
            # get block outputs
            for block_id, (args, kwargs) in enumerate(zip(total_block_args, total_block_kwargs)):
                out = block(*args, **kwargs)
                if isinstance(out, tuple):  # pragma: no cover
                    out = out[0]
                # update inputs of next block
                if len(total_block_args[block_id]) > 0:
                    total_block_args[block_id][0].copy_(out)
                elif "hidden_states" in total_block_kwargs[block_id]:
                    total_block_kwargs[block_id]["hidden_states"].copy_(out)
                cur_accelerator.synchronize()
            block = block.to("meta")  # clean device memory
            gc.collect()
            cur_accelerator.synchronize()
            block.load_state_dict(mapped_state_dict, assign=True)  # recover memory mapping
        return total_block_args, total_block_kwargs

    # Get non_persistent_buffers
    non_persistent_buffers = get_non_persistent_buffers(model)
    # Get prefix and number of attention blocks of transformer models.
    device = cur_accelerator.name()
    block_prefix, block_num = get_block_prefix(model)
    # Use prefix to fetch all attention blocks. For llama, it's LlamaDecoderLayer module.
    block_list = fetch_module(model, block_prefix)
    logger.info("Getting inputs of the first block")

    total_block_args, total_block_kwargs = get_block_kwargs(model, dataloader=dataloader, data=data, device=device)
    for i, block in enumerate(block_list):
        logger.info("Calibrating block: {}".format(i))
        # Calibrate current block and update args and kwargs for the next block.
        total_block_args, total_block_kwargs = calibrate_block_and_update_inputs(
            block, total_block_args, total_block_kwargs, device=device
        )
    # Load non_persistent_buffers
    load_non_persistent_buffers(model, non_persistent_buffers)
    # Clean host and device memory
    del total_block_args, total_block_kwargs
    gc.collect()
    cur_accelerator.synchronize()
    logger.info("Calibration is done!")



#
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Torch.nn.Module Class Definition."""
# Note: Do not import this file unless you have already imported torch,
# since the model classes inherit torch.nn.Module.
import math
from abc import abstractmethod

import numpy as np
import torch
from torch.autograd import Function
from torch.nn import functional as F

from neural_compressor.torch.utils import (
    accelerator,
    can_pack_with_numba,
    is_hpex_support_g_idx,
    logger,
)

from .utility import quant_tensor


class Matmul(torch.nn.Module):
    """Basic module for matmul."""

    def __init__(
        self,
    ) -> None:
        """Init the Matmul object."""
        super().__init__()

    def forward(self, X, Y):
        """Forward function."""
        return torch.matmul(X, Y)


class QDQLayer(torch.nn.Module):
    """Quantized and dequantized layer."""

    def __init__(self, module, input_scale=None) -> None:
        """Init the QDQLayer object."""
        super().__init__()
        self.quant = torch.ao.quantization.QuantStub()
        self.module = module
        self.dequant = torch.ao.quantization.DeQuantStub()
        self.input_scale = input_scale

    def forward(self, X):
        """Forward function."""
        if self.input_scale is not None:
            X = torch.mul(X, self.input_scale)
        X = self.quant(X)
        X = self.module(X)
        X = self.dequant(X)
        return X


class UnpackedWeightOnlyLinearParams(dict):
    """Contains all unpacked weight values."""

    def __init__(self, unpack_weight, scales, scale_bf16_to_fp8, unpack_zp, **kwargs):
        """Create dict."""
        super().__init__(
            int_weight=unpack_weight, scales=scales, scale_bf16_to_fp8=scale_bf16_to_fp8, zp=unpack_zp, **kwargs
        )

    def to(self, device):
        """Change device for all values."""
        for key, value in self.items():
            if isinstance(value, torch.Tensor) and value is not None:
                self[key] = value.to(device)
        return self


class WeightOnlyLinear(torch.nn.Module):
    """Base Weight Only Linear."""

    def __init__(
        self,
        in_features,
        out_features,
        dtype,
        bits,
        group_size,
        device,
        scale_dtype,
        **kwargs,
    ):
        """Initialization."""
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.dtype = dtype
        self.bits = bits
        self.group_size = group_size if group_size != -1 else in_features
        self.device = device
        self.scale_dtype = scale_dtype
        self.kwargs = kwargs
        self.enable_w4a8 = kwargs.get("enable_w4a8", False)

    def _post_init_for_w4a8(self):
        scale_dtype = self.scale_dtype
        self.act_scales = torch.nn.Parameter(torch.zeros((1,), dtype=scale_dtype), requires_grad=False)
        bf16_to_fp8_scales_shape = (1,)
        self.w_bf16_to_fp8_scale = torch.nn.Parameter(
            torch.zeros(bf16_to_fp8_scales_shape, dtype=scale_dtype), requires_grad=False
        )

    def post_init(self):
        """Initialization for W4A8 usage."""
        if self.enable_w4a8:
            self._post_init_for_w4a8()

    @abstractmethod
    def pack(self, *args, **kwargs):
        """Abstract method."""
        raise NotImplementedError("{} doesn't implement `pack` function. ".format(self.__class__.__name__))

    @abstractmethod
    def unpack(self, *args, **kwargs):
        """Abstract method."""
        raise NotImplementedError("{} doesn't implement `unpack` function. ".format(self.__class__.__name__))

    @abstractmethod
    def forward(self, input):
        """Abstract method."""
        raise NotImplementedError("{} doesn't implement `forward` function. ".format(self.__class__.__name__))

    def extra_repr(self) -> str:
        """Show message about WeighOnlyLinear."""
        tmp_str = "in_features={}, out_features={}, bits={}, group_size={}, bias={}".format(
            self.in_features,
            self.out_features,
            self.bits,
            self.group_size,
            self.bias is not None,
        )
        return tmp_str


class INCWeightOnlyLinear(WeightOnlyLinear):
    """INC Weight Only Linear."""

    def __init__(
        self,
        in_features,
        out_features,
        dtype="int",
        bits=4,
        group_size=32,
        zp=False,
        bias=False,
        scale_dtype=torch.float32,
        compression_dtype=torch.int32,
        compression_dim=1,
        g_idx=False,
        device="cpu",
        use_optimum_format=True,
        **kwargs,
    ):
        """Init the WeightOnlyLinear object.

        Args:
            in_features (int): input features.
            out_features (int): out features.
            dtype (str, optional):  the data type of the quantized model. Defaults to "int".
            bits (int, optional): number of bits for quantization. Defaults to 4.
            group_size (int, optional): size of the quantization group. Defaults to 32.
            zp (bool, optional): zero point. Defaults to False.
            bias (bool, optional): module bias. Defaults to False.
            scale_dtype (torch.Tensor, optional): the data type of quantization scale to be used.
                                                  Defaults to torch.float32.
            compression_dtype (torch.Tensor, optional): the target dtype after comoression.
                                                        Defaults to torch.int32.
            compression_dim (int, optional): select from [0, 1], 0 is output channel, 1 is input channel.
                                             Defaults to 1.
            g_idx (bool, optional): for recording the channel order.
            device (str, optional): choose device for compression. Defaults to cpu.
            use_optimum_format (bool, optional): use the popular huggingface compression format.
                1: compression_dim: weight = 1, zeros = 0 and both are transposed.
                2: zeros -= 1 before compression.
                3: g_idx: use same number for one group instead of recording the channel order.
                4. parameter name changed, such as 'packed_weight' -> 'qweight'.
                5. zeros is always needed even for sym.
        """
        super(INCWeightOnlyLinear, self).__init__(
            in_features,
            out_features,
            dtype,
            bits,
            group_size,
            device,
            scale_dtype=scale_dtype,
            **kwargs,
        )
        self.use_optimum_format = use_optimum_format
        if "int" not in self.dtype:  # for nf4, fp4
            from neural_compressor.torch.algorithms.weight_only.utility import FLOAT_MAPPING, INT_MAPPING

            self.use_optimum_format = False  # optimum_format doesn't suit for symmetric nf4 fp4.
            float_list = FLOAT_MAPPING[self.dtype]
            int_list = INT_MAPPING[self.dtype]
            self.int2float_mapping = {}
            for k, v in zip(int_list, float_list):
                self.int2float_mapping[k] = v
        self.compression_dim = compression_dim
        assert compression_dtype in [
            torch.int8,
            torch.int16,
            torch.int32,
            torch.int64,
        ], f"Only support torch.int8|16|32|64 as compressed dtype. but got {compression_dtype}"
        dtype_bits_mapping = {torch.int8: 8, torch.int16: 16, torch.int32: 32, torch.int64: 64}
        self.compress_bits = dtype_bits_mapping[compression_dtype]
        self.n_pack = self.compress_bits // self.bits
        # K is input channel, N is output channel
        assert compression_dim in [0, 1], (
            "Only support 0 or 1 as compression dimension, " + "0 is output channel, 1 is input channel."
        )
        self.register_buffer(
            "scale_bf16_to_fp8",
            torch.zeros(
                1,
                dtype=torch.bfloat16,
            ).to(device),
        )
        # scale_bf16_to_fp8 is only used in w4a8 measurement mode and currently supports only per-tensor scaling
        if self.use_optimum_format:
            self.float_type = torch.float16
            self.compression_dtype = torch.int32
            self.register_buffer(
                "scales",
                torch.zeros(
                    (math.ceil(in_features / self.group_size), out_features),
                    dtype=self.float_type,
                ).to(device),
            )
            self.register_buffer(
                "qweight",
                torch.zeros(
                    (math.ceil(in_features / self.n_pack), out_features),
                    dtype=self.compression_dtype,
                ).to(device),
            )
            self.register_buffer(
                "qzeros",
                torch.zeros(
                    (math.ceil(self.in_features / self.group_size), math.ceil(self.out_features / self.n_pack)),
                    dtype=self.compression_dtype,
                ).to(device),
            )
            self.register_buffer("bias", torch.zeros(self.out_features, dtype=self.float_type).to(device))
        else:
            self.compression_dtype = compression_dtype
            self.float_type = scale_dtype
            self.register_buffer(
                "scales",
                torch.zeros(
                    (out_features, math.ceil(in_features / self.group_size)),
                    dtype=self.float_type,
                ).to(device),
            )
            if compression_dim == 1:
                self.register_buffer(
                    "qweight",
                    torch.zeros(
                        (out_features, math.ceil(in_features / self.n_pack)),
                        dtype=self.compression_dtype,
                    ).to(device),
                )
                if zp:
                    self.register_buffer(
                        "qzeros",
                        torch.zeros(
                            (self.out_features, math.ceil(self.in_features / self.group_size / self.n_pack)),
                            dtype=self.compression_dtype,
                        ).to(device),
                    )
            else:
                self.register_buffer(
                    "qweight",
                    torch.zeros(
                        (math.ceil(out_features / self.n_pack), in_features),
                        dtype=self.compression_dtype,
                    ).to(device),
                )
                if zp:
                    self.register_buffer(
                        "qzeros",
                        torch.zeros(
                            (math.ceil(self.out_features / self.n_pack), math.ceil(self.in_features / self.group_size)),
                            dtype=self.compression_dtype,
                        ).to(device),
                    )
            if bias:
                self.register_buffer("bias", torch.zeros(self.out_features, dtype=self.float_type).to(device))
            else:
                self.bias = None
        if g_idx:
            self.register_buffer("g_idx", torch.zeros(in_features, dtype=torch.int32).to(device))
        else:
            self.g_idx = None
        self.post_init()

    def pack(self, int_weight, scales, zp, bias, scale_bf16_to_fp8=None, g_idx=None, **kwargs):
        """Pack int weight."""
        if self.use_optimum_format:
            self.scales = self.scales.T.contiguous()
            self.qweight = self.qweight.T.contiguous()
            self.qzeros = self.qzeros.T.contiguous()
        if int_weight.device.type != "meta":
            int_weight = int_weight.to(self.device)
        if self.use_optimum_format and zp is None:
            # to avoid overflow
            int_weight = int_weight.type(torch.int32)
            shift_bias = 2 ** (self.bits - 1)
            int_weight += shift_bias
            zp = torch.zeros_like(scales, dtype=torch.uint8) + shift_bias
        if bias is not None:
            assert hasattr(self, "bias"), "bias is not set when initializing."
            self.bias = bias.type(self.float_type).to(self.device)
        if g_idx is not None:
            assert hasattr(self, "g_idx"), "g_idx is not set when initializing."
            self.g_idx = g_idx.type(torch.int32).to(self.device)
            if self.use_optimum_format or "int" not in self.dtype:
                invperm = torch.argsort(self.g_idx)
                self.g_idx = invperm // self.group_size
                self.g_idx = self.g_idx.type(torch.int32).to(self.device)
        assert scales.shape == self.scales.shape, f"{scales.shape} != {self.scales.shape} Scale shape is mismatched."
        self.scales = scales.type(self.float_type).to(self.device)
        if scale_bf16_to_fp8 is not None:
            self.scale_bf16_to_fp8 = scale_bf16_to_fp8.type(self.float_type).to(self.device)
        if not self.use_optimum_format and self.compression_dim == 0:
            int_weight = int_weight.T.contiguous()
            self.qweight = self.qweight.T.contiguous()
        origin_shape = int_weight.shape
        target_shape = self.qweight.shape
        assert origin_shape[0] == target_shape[0], "output channels mismatch, please check."

        # pack weight
        self.qweight.copy_(self.pack_tensor(int_weight))
        if not self.use_optimum_format and self.compression_dim == 0:
            self.qweight = self.qweight.T.contiguous()

        if zp is not None:
            zp = zp.to(self.device)
            if self.use_optimum_format:
                zp -= 1
            if self.use_optimum_format or self.compression_dim == 0:
                zp = zp.T.contiguous()
                self.qzeros = self.qzeros.T.contiguous()
            assert hasattr(self, "qzeros"), "zp is not set when initializing."
            self.qzeros.copy_(self.pack_tensor(zp))
            if self.use_optimum_format or self.compression_dim == 0:
                self.qzeros = self.qzeros.T.contiguous()
        if self.use_optimum_format:
            self.scales = self.scales.T.contiguous()
            self.qweight = self.qweight.T.contiguous()
            self.qzeros = self.qzeros.T.contiguous()

    def unpack(self):
        """Unpack weight and zero point."""
        scales = self.scales.T.contiguous() if self.use_optimum_format else self.scales
        scale_bf16_to_fp8 = self.scale_bf16_to_fp8
        qweight = self.qweight.T.contiguous() if self.use_optimum_format else self.qweight

        device = scales.device
        # unpack weight
        if not self.use_optimum_format and self.compression_dim == 0:
            qweight = qweight.T.contiguous()
        weight = self.unpack_tensor(qweight)
        if not self.use_optimum_format and self.compression_dim == 0:
            weight = weight.T.contiguous()
        weight = weight[: self.out_features, : self.in_features]  # avoid oversize
        if "int" not in self.dtype:
            new_weight = torch.zeros(self.out_features, self.in_features).to(device)
            for k, v in self.int2float_mapping.items():
                new_weight += torch.where(weight == k, v, 0)
            weight = new_weight

        # unpack zero_point
        zp = None
        if hasattr(self, "qzeros"):
            qzeros = self.qzeros.T.contiguous() if self.use_optimum_format else self.qzeros
            if self.use_optimum_format or self.compression_dim == 0:
                qzeros = qzeros.T.contiguous()
            zp = self.unpack_tensor(qzeros)
            if self.use_optimum_format or self.compression_dim == 0:
                zp = zp.T.contiguous()
            zp = zp[: scales.shape[0], : scales.shape[1]]  # avoid oversize
            if self.use_optimum_format:
                # zp -= 1 may cause zp == -1, after recover it becomes 2**self.bits - 1
                zp += 1
                zp = torch.where(zp > (2**self.bits - 1), 0, zp)
        return UnpackedWeightOnlyLinearParams(weight, scales, scale_bf16_to_fp8, zp, g_idx=self.g_idx, bias=self.bias)

    def recover(self):
        """Recover fp32 weight from packed weight."""
        logger.debug(f"Recovering {self} weight")
        unpack_params_dict = self.unpack()
        weight = unpack_params_dict.get("int_weight")
        scales = unpack_params_dict.get("scales")
        scale_bf16_to_fp8 = unpack_params_dict.get("scale_bf16_to_fp8")
        zp = unpack_params_dict.get("zp")

        device = scales.device

        fp32_weight = torch.zeros(self.out_features, self.in_features, dtype=self.float_type).to(device)

        # recover fp32 weight
        if self.g_idx is None:
            # used for recovering fp32_weight
            self.g_idx = torch.tensor([i // self.group_size for i in range(self.in_features)], dtype=torch.int32).to(
                device
            )
        if zp is not None:
            # recover fp32 weight with int_weight, scale, and zero_point
            for idx in range(self.in_features):
                fp32_weight[:, idx] = (torch.subtract(weight[:, idx], zp[:, self.g_idx[idx]]).to(torch.int8)) * scales[
                    :, self.g_idx[idx]
                ]
        else:
            # recover fp32 weight with int_weight, scale
            for idx in range(self.in_features):
                fp32_weight[:, idx] = weight[:, idx] * scales[:, self.g_idx[idx]]

        return fp32_weight.to(scales.device)

    def pack_tensor_with_torch(self, raw_tensor):
        """Pack the tensor with torch.

        Args:
            raw_tensor (tensor): raw tensor.

        Returns:
            tensor: packed tensor.
        """
        target_len = math.ceil(raw_tensor.shape[1] / self.n_pack)
        packed_tensor = torch.zeros(raw_tensor.shape[0], target_len, dtype=self.compression_dtype).to(raw_tensor.device)
        mask = torch.tensor(2**self.bits - 1, dtype=self.compression_dtype).to(raw_tensor.device)
        for j in range(packed_tensor.shape[1]):
            start = self.n_pack * j
            end = self.n_pack * (j + 1)
            tmp = raw_tensor[:, start:end].type(self.compression_dtype)
            tmp &= mask
            for e in range(tmp.shape[1]):
                tmp[:, e] = tmp[:, e] << (self.bits * e)
                packed_tensor[:, j] |= tmp[:, e]
                accelerator.synchronize()
        return packed_tensor

    def unpack_tensor_with_torch(self, packed_tensor):
        """Unpack the tensor with torch.

        Args:
            packed_tensor (tensor): packed tensor.

        Returns:
            tensor: unpacked tensor.
        """
        target_dtype = torch.int16
        target_len = packed_tensor.shape[1] * self.n_pack
        unpacked_tensor = torch.zeros(packed_tensor.shape[0], target_len, dtype=target_dtype).to(packed_tensor.device)
        mask = torch.tensor(2**self.bits - 1, dtype=self.compression_dtype).to(packed_tensor.device)
        for j in range(packed_tensor.shape[1]):
            for e in range(self.n_pack):
                index = j * self.n_pack + e
                tmp = packed_tensor[:, j]
                tmp = tmp << (self.compress_bits - self.bits * (e + 1))
                tmp = tmp >> self.compress_bits - self.bits
                if hasattr(self, "qzeros"):
                    tmp &= mask  # remove sign bit
                unpacked_tensor[:, index].copy_(tmp.type(target_dtype))
                accelerator.synchronize()
        return unpacked_tensor

    def pack_array_with_numba(
        self, raw_array: np.ndarray, n_pack: int, bits: int, compress_bits: int, compression_dtype=np.int32
    ) -> np.ndarray:
        """Packs the input array by combining elements into a specified bit-width format using NumPy.

        Args:
            raw_array (np.ndarray): The array to be packed. Shape: [out_features, in_features] or [1, in_features].
            n_pack (int): The number of elements to be packed together.
            bits (int): The number of bits for each element.
            compress_bits (int): The number of bits for each element of the compressed array, supported 2, 4, 8.
            compression_dtype (np.dtype, optional): The data type of the compressed array. Defaults to np.int32.

        Returns:
            np.ndarray: The packed array.
        """
        # Try to pack with numba to accelerate the packing process.
        # If numba is not availabll or the packing method is not supported,
        # fallback to the torch implementation.
        if not can_pack_with_numba():
            return self.pack_tensor_with_torch(torch.from_numpy(raw_array)).cpu().numpy()
        from neural_compressor.torch.utils.bit_packer import bit_packers

        pack_func_name = (bits, compress_bits)
        if pack_func_name not in bit_packers:
            logger.warning(
                f"Unsupported packing with bits: {bits}, compress_bits: {compress_bits} using numba, fallback to torch implementation."
            )
            return self.pack_tensor_with_torch(torch.from_numpy(raw_array)).cpu().numpy()
        out_features, in_features = raw_array.shape
        new_in_features = (in_features + n_pack - 1) // n_pack
        packed_array = np.zeros((out_features, new_in_features), dtype=compression_dtype)
        raw_array = raw_array.astype(compression_dtype)
        pack_method = bit_packers[pack_func_name]
        return pack_method(raw_array, packed_array, n_pack, new_in_features)

    def pack_tensor_with_numpy_impl(self, raw_tensor):
        """The implement of packing tensor with numpy."""
        raw_array = raw_tensor.cpu().numpy()
        target_len = np.ceil(raw_array.shape[1] / self.n_pack).astype(int)
        target_dtype = torch.tensor(0, dtype=self.compression_dtype).numpy().dtype
        packed_array = np.zeros((raw_array.shape[0], target_len), dtype=target_dtype)
        mask = np.uint8(2**self.bits - 1)
        for j in range(packed_array.shape[1]):
            start = self.n_pack * j
            end = self.n_pack * (j + 1)
            tmp = raw_array[:, start:end].astype(target_dtype)
            tmp &= mask
            for e in range(tmp.shape[1]):
                tmp[:, e] = np.left_shift(tmp[:, e], self.bits * e)
                packed_array[:, j] |= tmp[:, e]
        packed_tensor = torch.from_numpy(packed_array).to(device=raw_tensor.device)
        return packed_tensor

    def pack_tensor_with_numpy(self, raw_tensor):
        """Pack the tensor with numpy."""
        if self.bits == 8 and self.compression_dtype == torch.int8:
            return raw_tensor
        if self.bits not in [2, 4, 8]:
            return self.pack_tensor_with_numpy_impl(raw_tensor)
        compression_dtype = torch.tensor(0, dtype=self.compression_dtype).numpy().dtype
        packed_array = self.pack_array_with_numba(
            raw_tensor.cpu().numpy(), self.n_pack, self.bits, self.compress_bits, compression_dtype
        )
        return torch.from_numpy(packed_array).to(device=raw_tensor.device)

    def unpack_tensor_with_numpy(self, packed_tensor):
        """Unpack the packed tensor with numpy."""
        packed_array = packed_tensor.cpu().numpy()
        target_dtype = np.int16
        if self.bits == 8 and self.compression_dtype == torch.int8 and hasattr(self, "qzeros"):
            # special case for unpacking uint8 date from int8 compression_dtype
            target_dtype = np.uint8
        target_len = packed_array.shape[1] * self.n_pack
        unpacked_array = np.zeros((packed_array.shape[0], target_len), dtype=target_dtype)
        mask = np.uint8(2**self.bits - 1)
        for j in range(packed_array.shape[1]):
            for e in range(self.n_pack):
                index = j * self.n_pack + e
                tmp = packed_array[:, j]
                tmp = np.left_shift(tmp, self.compress_bits - self.bits * (e + 1))
                tmp = np.right_shift(tmp, self.compress_bits - self.bits)
                if hasattr(self, "qzeros"):
                    tmp &= mask
                unpacked_array[:, index] = tmp.astype(target_dtype)
        unpacked_tensor = torch.from_numpy(unpacked_array).to(device=packed_tensor.device)
        return unpacked_tensor

    def pack_tensor(self, raw_tensor):
        """Pack tensor."""
        if "cuda" in raw_tensor.device.type:
            return self.pack_tensor_with_torch(raw_tensor)
        else:
            return self.pack_tensor_with_numpy(raw_tensor)

    def unpack_tensor(self, packed_tensor):
        """Unpack tensor."""
        if "cuda" in packed_tensor.device.type:
            return self.unpack_tensor_with_torch(packed_tensor)
        else:
            return self.unpack_tensor_with_numpy(packed_tensor)

    def forward(self, input):
        """Forward function."""
        if not hasattr(self, "weight"):
            weight = self.recover()
            device = self.scales.device
            if weight.dtype == torch.float16 and device.type == "cpu":
                weight = weight.float()
                self.bias = self.bias.float() if self.bias is not None else None
        if True:  # keep reusing self.weight due to recover is too slow.
            if not hasattr(self, "weight"):
                self.weight = weight
            input = input.type(self.weight.dtype)
            logger.debug(f"Calculating {self}")
            return F.linear(input, self.weight, self.bias)
        else:
            input = input.type(weight.dtype)
            return F.linear(input, weight, self.bias)

    def extra_repr(self) -> str:
        """Extract the configuration string.

        Returns:
            str: the configuration string.
        """
        tmp_str = "in_features={}, out_features={}, bits={}, group_size={}, bias={}".format(
            self.in_features,
            self.out_features,
            self.bits,
            self.group_size,
            self.bias is not None,
        )
        if self.use_optimum_format:
            tmp_str += ", use_optimum_format=True"
        return tmp_str


class HPUWeightOnlyLinear(WeightOnlyLinear):
    """Weight Only Linear for HPU device."""

    def __init__(
        self,
        in_features,
        out_features,
        dtype="int",
        bits=4,
        group_size=32,
        zp=False,
        bias=False,
        scale_dtype=torch.bfloat16,
        compression_dtype=torch.int32,
        compression_dim=1,
        g_idx=False,
        device="hpu",
        use_optimum_format=True,
        **kwargs,
    ):
        """Init the WeightOnlyLinear object.

        Args:
            in_features (int): input features.
            out_features (int): out features.
            dtype (str, optional):  the data type of the quantized model. Defaults to "int".
            bits (int, optional): number of bits for quantization. Defaults to 4.
            group_size (int, optional): size of the quantization group. Defaults to 32.
            zp (bool, optional): zero point. Defaults to False.
            bias (bool, optional): module bias. Defaults to False.
            scale_dtype (torch.Tensor, optional): the data type of quantization scale to be used.
                                                  Defaults to torch.float32.
            compression_dtype (torch.Tensor, optional): the target dtype after comoression.
                                                        Defaults to torch.int32.
            compression_dim (int, optional): select from [0, 1], 0 is output channel, 1 is input channel.
                                             Defaults to 1.
            g_idx (bool, optional): for recording the channel order.
            device (str, optional): choose device for compression. Defaults to cpu.
            use_optimum_format (bool, optional): use the popular huggingface compression format.
                1: compression_dim: weight = 1, zeros = 0 and both are transposed.
                2: zeros -= 1 before compression.
                3: g_idx: use same number for one group instead of recording the channel order.
                4. parameter name changed, such as 'packed_weight' -> 'qweight'.
                5. zeros is always needed even for sym.
        """
        super(HPUWeightOnlyLinear, self).__init__(
            in_features,
            out_features,
            dtype,
            bits,
            group_size,
            device,
            scale_dtype=scale_dtype,
            **kwargs,
        )
        self.float_type = torch.bfloat16
        self.compression_dim = compression_dim
        self.compression_dtype = compression_dtype

        if bits != 4:
            raise NotImplementedError("Only 4 bits are supported.")
        self.maxq = 2**self.bits - 1

        if bias:
            self.register_buffer("bias", torch.zeros(self.out_features, dtype=self.float_type).to(self.device))
        else:
            self.bias = None

        self.register_buffer(
            "qweight",
            torch.zeros((in_features, out_features // 32 * self.bits), dtype=self.compression_dtype).to(self.device),
        )

        self.register_buffer(
            "qzeros",
            torch.zeros(
                (
                    math.ceil(in_features / self.group_size),
                    out_features // 32 * self.bits,
                ),
                dtype=self.compression_dtype,
            ),
        )
        self.register_buffer(
            "scales",
            torch.zeros(
                (math.ceil(in_features / self.group_size), out_features),
                dtype=self.float_type,
            ),
        )
        self.register_buffer(
            "scale_bf16_to_fp8",
            torch.zeros(
                1,
                dtype=self.float_type,
            ),
        )
        if g_idx:
            self.register_buffer(
                "g_idx",
                torch.tensor([i // self.group_size for i in range(in_features)], dtype=torch.int32),
            )
        else:
            self.g_idx = None
        self.support_g_idx = is_hpex_support_g_idx()

        self.half_indim = self.in_features // 2

        self.wf = torch.tensor(list(range(0, 32, self.bits)), dtype=torch.int32).unsqueeze(0)
        self.matmul_internal = Matmul()
        self.post_init()

    def forward(self, input):
        """The forward function of HPUWeighOnlyLinear."""
        input_dtype = input.dtype
        output_shape = input.shape[:-1] + (self.out_features,)
        scales = self.scales
        qweight = self.qweight
        zeros = self.qzeros
        g_idx = self.g_idx
        if self.support_g_idx:
            weight = torch.ops.hpu.convert_from_uint4(qweight, scales, zeros, input_dtype, g_idx)
        else:
            # The released docker 1.22.0 only supports 4 arguments
            weight = torch.ops.hpu.convert_from_uint4(qweight, scales, zeros, input_dtype)
        output = self.matmul_internal(input, weight)
        output = output.to(dtype=input_dtype).reshape(
            output_shape
        )  # A cast is needed here as for some reason the vecquant2matmul_faster_old still allocate a float32 output.
        output = output + self.bias if self.bias is not None else output
        return output

    # [SW-234528]: This is a temporary workaround; remove it when find a better solution.
    @staticmethod
    def is_g_idx_ordered(g_idx: torch.Tensor, group_size: int) -> bool:
        """Check if g_idx is ordered according to group_size."""
        if group_size == -1:
            return True
        if g_idx.numel() % group_size != 0:
            raise ValueError("The length of g_idx must be divisible by group_size.")

        for i in range(0, g_idx.numel(), group_size):
            if len(torch.unique(g_idx[i : i + group_size])) != 1:
                return False
        return True

    def pack(self, int_weight, scales, zp, scale_bf16_to_fp8=None, bias=None, g_idx=None):
        """Pack weight and zero point."""
        logger.debug("Packing for HPU")

        scales = scales.T.contiguous()
        qzeros = zp.T.contiguous()
        qweight = int_weight.T.contiguous()

        self.scales = scales.to(dtype=torch.bfloat16)
        self.scale_bf16_to_fp8 = scale_bf16_to_fp8.to(dtype=torch.bfloat16)

        # weights and zp are on device from unpack, need to load to cpu for packing
        self.qweight = qweight.cpu()
        new_qweight = self.pack_tensor(self.qweight)
        self.qweight = new_qweight.to("hpu")

        self.qzeros = qzeros.cpu()
        new_qzeros = self.pack_tensor(self.qzeros)
        self.qzeros = new_qzeros.to("hpu")

        if bias is not None:
            self.bias = bias.to("hpu").to(torch.bfloat16)

        if g_idx is not None:
            self.g_idx = g_idx.to("hpu").to(torch.int32)

    def unpack(self):
        """Unpack weight and zero point."""
        logger.debug("Unpacking from HPU")
        self.qweight = self.qweight.cpu()
        weight = torch.bitwise_right_shift(
            torch.unsqueeze(self.qweight, 1).expand(-1, 32 // self.bits, -1),
            self.wf.unsqueeze(-1),
        ).to(torch.int16 if self.bits == 8 else torch.int8)
        weight = torch.bitwise_and(weight, (2**self.bits) - 1)
        weight = weight.reshape((weight.shape[0] * weight.shape[1], weight.shape[2]))
        self.qweight = self.qweight.to(self.device)

        zeros = torch.bitwise_right_shift(
            torch.unsqueeze(self.qzeros, 2).expand(-1, -1, 32 // self.bits),
            self.wf.unsqueeze(0),
        ).to(torch.int16 if self.bits == 8 else torch.int8)

        zeros = torch.bitwise_and(zeros, (2**self.bits) - 1).to(
            self.scales.dtype
        )  # NOTE: It appears that casting here after the `zeros = zeros + 1` is important.
        zeros = zeros + 1
        zeros = zeros.reshape(-1, 1, zeros.shape[1] * zeros.shape[2])
        return weight, zeros

    def pack_tensor(self, input, bits=4):
        """Pack tensor."""
        normal = input.to(torch.int32)
        q = torch.zeros((normal.shape[0], normal.shape[1] // 32 * bits), dtype=torch.int32)
        i = 0
        col = 0
        while col < q.shape[1]:
            for j in range(i, i + (32 // bits)):
                q[:, col] |= normal[:, j] << (bits * (j - i))
            i += 32 // bits
            col += 1
        q = q.to(torch.int32)
        return q


class FakeAffineTensorQuantFunction(Function):
    """Fake version of affine quantization."""

    @staticmethod
    def forward(ctx, inputs, num_bits=4, group_size=1024, scheme="asym"):
        """As it will be only applied on activation with per tensor granularity, broadcast is not needed.

        Args:
            ctx: Pytorch convention.
            inputs: A Tensor of type float32.
            min_range: A float.
            max_range: A float.
            num_bits: An integer

        Returns:
            outputs: A Tensor of type output_dtype
        """
        return quant_tensor(inputs, num_bits, group_size, scheme)

    @staticmethod
    def backward(ctx, grad_outputs):
        """Backward function.

        Args:
            ctx: Pytorch convention.
            grad_output: A tensor of gradient of outputs

        Returns:
            grad_inputs: A tensor of gradient
        """
        return grad_outputs, None, None, None


class TEQLinearFakeQuant(torch.nn.Module):
    """Wrapper quantization linear."""

    def __init__(self, orig_layer, alpha=None, num_bits=4, group_size=-1, scheme="asym"):
        """A forward hook to linear module.

        Args:
            orig_layer: the original module
            alpha: trainable alpha/scale
            num_bits: quantization level
            group_size: for fine-grained quantization.
            scheme: symmetric quantization or asymmetric quantization.
        """
        super(TEQLinearFakeQuant, self).__init__()
        self.orig_layer = orig_layer
        self.alpha = alpha

        self.num_bits = num_bits
        self.group_size = group_size
        self.scheme = scheme

    def forward(self, x):
        """Forward function."""
        alpha = torch.clip(self.alpha, 1e-5)
        shape_len = len(x.shape) - 1
        shape = (1,) * shape_len + (-1,)
        x = x / alpha.view(shape)
        weight = self.orig_layer.weight
        weight = weight * alpha.unsqueeze(dim=0)
        weight_q = FakeAffineTensorQuantFunction().apply(weight, self.num_bits, self.group_size, self.scheme)
        return F.linear(x, weight_q, self.orig_layer.bias)


class MulLinear(torch.nn.Module):
    """Linear wrapper to apply scale to input."""

    def __init__(self, module, input_scale=None):
        """A forward hook to save input max of a module.

        Args:
            module: the linear module.
            input_scale: scale for input.
        """
        super().__init__()
        if input_scale is None:
            input_scale = torch.empty(module.in_features)
        self.register_buffer("input_scale", input_scale)
        self.add_module("linear", module)

    @property
    def weight(self):
        """Property weight."""
        return self.linear.weight

    @weight.setter
    def weight(self, weight):
        """Property weight setter."""
        self.linear.weight = weight

    def forward(self, X):
        """Forward function."""
        X = torch.mul(X, self.input_scale)
        X = self.linear(X)
        return X

    def _update_linear(self):
        # update linear weight with input_scale
        scale = self.input_scale.view(1, self.input_scale.shape[0])
        with torch.no_grad():
            self.linear.weight /= scale

    def _recover_linear(self):
        # remove mul and reset sq_linear for ipex inference
        scale = self.input_scale.view(1, self.input_scale.shape[0])
        with torch.no_grad():
            self.linear.weight *= scale



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


try:
    import habana_frameworks.torch.hpex  # pylint: disable=E0401

    _hpex_available = True
except:
    _hpex_available = False

from neural_compressor.common import logger


class LMEvalParser:
    def __init__(
        self,
        model="hf",
        tasks="lambada_openai",
        model_args="",
        user_model=None,
        tokenizer=None,
        num_fewshot=None,
        batch_size=1,
        max_batch_size=None,
        device=None,
        output_path=None,
        limit=None,
        use_cache=None,
        cache_requests=None,
        check_integrity=False,
        write_out=False,
        log_samples=False,
        show_config=False,
        include_path=None,
        gen_kwargs=None,
        verbosity="INFO",
        wandb_args="",
        predict_only=False,
        seed=[0, 1234, 1234],
        trust_remote_code=False,
        pad_to_buckets=None,  # used by HPU to align input length for performance.
        buckets=[32, 64, 128, 256, 512, 1024, 2048, 4096],  # used by HPU to limit input length range.
        add_bos_token=False,
    ):
        self.model = model
        self.tasks = tasks
        self.model_args = model_args
        self.user_model = user_model
        self.tokenizer = tokenizer
        self.num_fewshot = num_fewshot
        self.batch_size = batch_size
        self.max_batch_size = max_batch_size
        self.device = device
        self.output_path = output_path
        self.limit = limit
        self.use_cache = use_cache
        self.cache_requests = cache_requests
        self.check_integrity = check_integrity
        self.write_out = write_out
        self.log_samples = log_samples
        self.show_config = show_config
        self.include_path = include_path
        self.gen_kwargs = gen_kwargs
        self.verbosity = verbosity
        self.wandb_args = wandb_args
        self.predict_only = predict_only
        self.seed = seed
        self.trust_remote_code = trust_remote_code
        if pad_to_buckets is None:
            if _hpex_available:
                self.pad_to_buckets = True
            else:
                self.pad_to_buckets = False
        else:
            self.pad_to_buckets = pad_to_buckets
        self.buckets = buckets
        self.add_bos_token = add_bos_token
        self._post_init()

    def _check_add_bos_token(self):
        if not self.add_bos_token:
            logger.warning(
                (
                    "`add_bos_token` is set to False. "
                    "If the model was trained or fine-tuned with a BOS token, this may lead to incorrect results."
                )
            )

    def _post_init(self):
        self._check_add_bos_token()



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor Pytorch quantization AutoTune API."""


from copy import deepcopy
from typing import Callable, List, Optional, Union

import torch

from neural_compressor.common.base_config import BaseConfig, get_all_config_set_from_config_registry
from neural_compressor.common.base_tuning import EvaluationFuncWrapper, TuningConfig, init_tuning
from neural_compressor.common.utils import dump_elapsed_time
from neural_compressor.torch.quantization import quantize
from neural_compressor.torch.quantization.config import FRAMEWORK_NAME, AutoRoundConfig, RTNConfig
from neural_compressor.torch.utils import constants, logger

__all__ = [
    "autotune",
    "get_all_config_set",
    "get_rtn_double_quant_config_set",
]


def get_rtn_double_quant_config_set() -> List[RTNConfig]:
    """Generate RTN double quant config set.

    Returns:
        List[RTNConfig]: a set of quant config
    """
    rtn_double_quant_config_set = []
    for double_quant_type, double_quant_config in constants.DOUBLE_QUANT_CONFIGS.items():
        rtn_double_quant_config_set.append(RTNConfig.from_dict(double_quant_config))
    return rtn_double_quant_config_set


def get_all_config_set() -> Union[BaseConfig, List[BaseConfig]]:
    """Generate all quant config set.

    Returns:
        Union[BaseConfig, List[BaseConfig]]: a set of quant config
    """
    return get_all_config_set_from_config_registry(fwk_name=FRAMEWORK_NAME)


def _deepcopy_warp(model):
    """Create a deep copy of the model while preserving specific attributes.

    Args:
        model (torch.nn.Module): The model to deep copy.

    Returns:
        torch.nn.Module: A deep copy of the model with preserved attributes.
    """
    additional_attr_lst = ["_exported", "dynamic_shapes"]
    original_attr = {key: getattr(model, key, None) for key in additional_attr_lst}
    new_model = deepcopy(model)
    for key, value in original_attr.items():
        setattr(new_model, key, value)
    return new_model


def _preprocess_model_quant_config(model, quant_config):
    """Preprocess model and quant config before quantization.

    Args:
        model (torch.nn.Module): The model to be quantized.
        quant_config (TuningConfig): The quantization configuration to preprocess.

    Returns:
        Tuple[torch.nn.Module, TuningConfig]: The preprocessed model and quantization configuration.
    """
    for config in quant_config.config_set:
        # handle tokenizer attribute in AutoRoundConfig
        if isinstance(config, AutoRoundConfig):
            _tokenizer_backup = getattr(config, "tokenizer", None)
            if _tokenizer_backup is not None:
                setattr(model, "tokenizer", _tokenizer_backup)
                delattr(config, "tokenizer")
    return model, quant_config


@dump_elapsed_time("Pass auto-tune")
def autotune(
    model: torch.nn.Module,
    tune_config: TuningConfig,
    eval_fn: Callable,
    eval_args=None,
    run_fn=None,
    run_args=None,
    example_inputs=None,
):
    """The main entry of auto-tune.

    Args:
        model (torch.nn.Module): The model to be quantized.
        tune_config (TuningConfig): The configuration for the auto-tuning process.
        eval_fn (Callable): for evaluation of quantized models.
        eval_args (tuple, optional): arguments used by eval_fn. Defaults to None.
        run_fn (Callable, optional): for calibration to quantize model. Defaults to None.
        run_args (tuple, optional): arguments used by run_fn. Defaults to None.
        example_inputs (tensor/tuple/dict, optional): used to trace torch model. Defaults to None.

    Returns:
        The quantized model.
    """
    best_quant_model = None
    model, tune_config = _preprocess_model_quant_config(model, tune_config)
    eval_func_wrapper = EvaluationFuncWrapper(eval_fn, eval_args)
    config_loader, tuning_logger, tuning_monitor = init_tuning(tuning_config=tune_config)
    baseline: float = eval_func_wrapper.evaluate(_deepcopy_warp(model))
    tuning_monitor.set_baseline(baseline)
    tuning_logger.tuning_start()
    for trial_index, quant_config in enumerate(config_loader, 1):
        tuning_logger.trial_start(trial_index=trial_index)
        tuning_logger.execution_start()
        logger.info(quant_config.to_dict())
        # !!! Make sure to use deepcopy only when inplace is set to `True`.
        q_model = quantize(
            _deepcopy_warp(model),
            quant_config=quant_config,
            run_fn=run_fn,
            run_args=run_args,
            inplace=True,
            example_inputs=example_inputs,
        )
        tuning_logger.execution_end()
        tuning_logger.evaluation_start()
        eval_result: float = eval_func_wrapper.evaluate(q_model)
        tuning_logger.evaluation_end()
        tuning_monitor.add_trial_result(trial_index, eval_result, quant_config)
        tuning_logger.trial_end(trial_index)
        if tuning_monitor.need_stop():
            logger.info("Stopped tuning.")
            best_trial_record = tuning_monitor.get_best_trial_record()
            if best_trial_record.trial_index != trial_index:
                logger.info("Re-quantizing with best quantization config...")
                del q_model  # maybe gc.collect() is needed for memory release
                best_quant_config: BaseConfig = best_trial_record.quant_config
                # !!! Make sure to use deepcopy only when inplace is set to `True`.
                q_model = quantize(
                    _deepcopy_warp(model),
                    quant_config=best_quant_config,
                    run_fn=run_fn,
                    run_args=run_args,
                    inplace=True,
                    example_inputs=example_inputs,
                )
            best_quant_model = q_model  # quantize model inplace
            break
    tuning_logger.tuning_end()
    return best_quant_model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch

from abc import abstractmethod

from .._quant_common.quant_config import MeasureExclude, QuantMode, get_hqt_config, set_hqt_config
from .save_measure import gmod_list
from .scale_methods.scale_method_config import ScaleMethodString
from ..utils.logger import logger
from .common import load_file, save_file, ShapeList
from .patching_common import generate_model_info, mod_default_dict, mod_types, parent_child_mod_dict
from ..model_configs import ModuleExtraConfig
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator
from neural_compressor.torch.algorithms.fp8_quant.model_configs import (
    OBSERVER_TYPES,
    OBSERVER_PARAMS,
    IMOD_DICT,
)
from neural_compressor.torch.algorithms.fp8_quant._core.common import dequant_original_fp8_weight_if_needed


cur_accelerator = auto_detect_accelerator()


def patch_module_measure(mod, mconfig, mod_dict):
    """Replaces the module with patched module according to mconfig.

    Args:
        mod (nn.module): The module that will be replaced with patched module that measures the inputs.
        mconfig (e.g. MaxAbsObserver/MaxAbsPerChannelObserver): The observer object that will measure the parameters.
        mod_dict (dict): dictionary from module name to its patched module.

    Returns:
        nn.module: The new module after patching.
    """
    parent = parent_child_mod_dict[mod].parent
    name = parent_child_mod_dict[mod].name
    parent = parent_child_mod_dict[mod].parent
    patched_mod = mod_dict[mod.__class__.__name__].patched_module(mod, parent, mconfig, name)
    setattr(parent, name, patched_mod)
    return patched_mod


def init_measure_object(mod, name, observer_class, mod_type, skip_measure_output, d_shape=None, params=None):
    input_observer = [
        observer_class(
            "input%d" % (i),
            mod,
            None if d_shape is None else d_shape.inputs[i],
            params=None if params is None else params.inputs[i],
        )
        for i in range(mod_type.num_inputs)
    ]
    if (not mod_type.required_output) and skip_measure_output:
        # excluding output measurements
        output_observer = None
    else:
        output_observer = [
            observer_class(
                "output%d" % (i),
                mod,
                None if d_shape is None else d_shape.outputs,
                params=None if params is None else params.outputs,
            )
            for i in range(mod_type.num_outputs)
        ]
    params_observer = {
        param_name: observer_class(
            param_name,
            mod,
            None if d_shape is None else d_shape.params[param_name],
            params=None if params is None else params.params[param_name],
        )
        for param_name in mod_type.param_names
    }
    measure_object = ModuleExtraConfig(input_observer, output_observer, params_observer, None, params)
    return measure_object


def prepare_model(model, mod_list=None):
    """Defines the observer class and modules for measurement as preparation.

    Args:
        model (nn.module): The model that will be measured.
        mod_list (list, optional): The specific submodules that will be measured in the model. Defaults to None.
    """
    config = get_hqt_config(model).cfg
    observer_class = OBSERVER_TYPES[config["observer"]]
    if (config.get("shape_file", None) is not None) and (observer_class != OBSERVER_TYPES["shape"]):
        shapes_fname = config["shape_file"] + ".json"
        d_shapes = load_file(shapes_fname, ShapeList, False)
    else:
        d_shapes = None
    # TODO SW-220992: Add dependency check for observer, just like maxabs_per_channel is dependent on shape
    if not d_shapes and observer_class == OBSERVER_TYPES["maxabs_per_channel"]:
        raise RuntimeError("Required shape files are missing from measurement directory")
    gmod_list.extend(mod_list)
    generate_model_info(model)
    register_patched_measure_modules(model, mod_list, observer_class, d_shapes)

def setup_calibration_counter(model, config):
    # used for automatically dumping measurements
    calibration_sample_interval = int(config["calibration_sample_interval"])
    if calibration_sample_interval > 0:
        from .save_measure import add_calibration_samples_counter
        add_calibration_samples_counter(model, calibration_sample_interval)

def register_patched_measure_modules(model, mod_list, observer_class, d_shapes=None):
    """Replace the submodules of the model that appear in mod_list with a patched submodule that uses the given observer_class
    so the submodule will perform measurement on inputs/outputs in forward stage.
    Weights measurement is done during model preparation as they are static.

    Args:
        model (nn.module): The model that will be measured.
        mod_list (list): The specific submodules that will be measured in the model.
        observer_class (e.g. MaxAbsObserver/MaxAbsPerChannelObserver): The observer type that will measure the weights.
        d_shapes (dict, optional): Defaults to None.
    """
    top_level_config = get_hqt_config(model)
    config = top_level_config.cfg
    setup_calibration_counter(model, config)
    skip_outputs_measurements = config["measure_exclude"] & (MeasureExclude.OUTPUT | MeasureExclude.ALL)
    patched_types = set()
    non_patched_types = set()
    patched_modules = []
    with torch.no_grad():
        for name, mod in model.named_modules():
            if (name in mod_list) or (mod_list is None):
                IMOD_DICT[mod] = name
                mod_type_str = mod.__class__.__name__
                mod_type = config["mod_dict"][mod_type_str]
                params = (
                    OBSERVER_PARAMS[config["observer"]][mod_type]
                    if (config["observer"] in OBSERVER_PARAMS) and (mod_type in OBSERVER_PARAMS[config["observer"]])
                    else None
                )
                patched_types.add(type(mod))

                set_hqt_config(mod, top_level_config)  # set config in the module, as it consumed by the patched module
                if mod_type == "dynamic_moe" and hasattr(mod, "num_experts"):
                    # override default number of outputs for dynamic moe
                    mod_types[mod_type].num_outputs = mod.num_experts+1
                    logger.warning(f"Dynamic moe num_outputs set to {mod.num_experts+1}")
                mod_extra_config = (
                    init_measure_object(
                        mod,
                        name,
                        observer_class,
                        mod_types[mod_type],
                        skip_outputs_measurements,
                        (d_shapes[name] if ((d_shapes is not None) and (name in d_shapes)) else None),
                        params,
                    )
                    if mod_default_dict[mod_type_str].should_measure_and_quant
                    else None
                )
                pmod = patch_module_measure(mod, mod_extra_config, mod_default_dict)
                if pmod._mod_extra_config:
                    for param_name in pmod._mod_extra_config.params:
                        param = getattr(pmod, param_name)
                        param = dequant_original_fp8_weight_if_needed(pmod.orig_mod, param)
                        if config["measure_on_hpu"]:
                            param = param.to(cur_accelerator.name())
                        pmod._mod_extra_config.params[param_name].measure(param)
                        cur_accelerator.synchronize()
                if observer_class == OBSERVER_TYPES["save"]:
                    save_module(pmod)
                patched_modules.append(name)
            else:
                non_patched_types.add(type(mod))
    logger.debug("Patched module types: %s", patched_types)
    logger.debug("None-patched module types: %s", non_patched_types)
    logger.debug("Patched modules: %s", patched_modules)
    logger.debug("Total patched modules: %d", len(patched_modules))
    if config["measure_on_hpu"]:
        model = model.to(cur_accelerator.name())
    cur_accelerator.synchronize()


def load_measurements(model, fname):
    config = get_hqt_config(model).cfg
    source_fname = fname if fname is not None else config["measure_file"]
    fname_np = source_fname + ".npz"
    d = load_file(
        fname_np,
        np.ndarray,
        fail_on_file_not_exist=(config["scale_method"] != ScaleMethodString.UNIT_SCALE),
    )
    from collections import defaultdict

    d = defaultdict(lambda: None, d)

    return d


def save_json(d, fname):
    with open(fname, "w") as f:
        json.dump(d, f, indent=4)


def load_json(fname):
    with open(fname, "r") as f:
        d = json.load(f)
    return d


def save_module(mod):
    folder_name = os.path.join(mod.config["dump_stats_base_path"], "tensors")
    os.makedirs(folder_name, exist_ok=True)
    file_base_name = os.path.join(folder_name, IMOD_DICT[mod] + "_module.pt")
    torch.save(mod.state_dict(), file_base_name)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
import torch
from torch.ao.quantization.fx._decomposed import quantized_decomposed_lib
from abc import abstractmethod
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator
from .quantized_func_wrappers import get_quantized_func_wrapper, OP_TYPE


cur_accelerator = auto_detect_accelerator()

from .._core.scale_handler import add_scale_registry, get_scale_dtype
from .._quant_common.quant_config import ScaleFormat
from .common import QuantTensorType
from .scale_handler import create_scale_tensor


class QuantDequantBase(nn.Module):
    def __init__(self, lp_dtype, hp_dtype="", *args, **kwargs):
        super().__init__()  # Initialize nn.Module
        add_scale_registry(self)
        self.lp_dtype = lp_dtype
        self.hp_dtype = hp_dtype
        self.scale_format = kwargs.get("scale_format", ScaleFormat.CONST)
        self.use_qdq = kwargs.get("use_qdq", False)
        if self.use_qdq:
            self.qdq_init()

    def qdq_init(self):
        self.quant_min = int(torch.finfo(self.lp_dtype).min)
        self.quant_max = int(torch.finfo(self.lp_dtype).max)

        if cur_accelerator.current_device_name() == "hpu":
            import habana_frameworks.torch.utils.experimental as htexp
            if htexp._get_device_type() == htexp.synDeviceType.synDeviceGaudi2 and self.lp_dtype == torch.float8_e4m3fn:
                self.quant_min = int(torch.finfo(torch.float8_e4m3fnuz).min)
                self.quant_max = int(torch.finfo(torch.float8_e4m3fnuz).max)

        if self.scale_format == ScaleFormat.CONST:
            self.zero_point = nn.Parameter(torch.tensor(0.))
        else:
            self.zero_point = 0
        self.forward = self.forward_qdq

    @abstractmethod
    def forward(self, *args, **kwargs):
        pass

    @abstractmethod
    def forward_qdq(self, *args, **kwargs):
        pass

    def extra_repr(self) -> str:
        return f"lp_dtype={self.lp_dtype}, hp_dtype={self.hp_dtype}"


class QuantDequantNone(QuantDequantBase):
    def __init__(self, lp_dtype, hp_dtype, *args, **kwargs):
        super(QuantDequantNone, self).__init__(lp_dtype, hp_dtype, *args, **kwargs)

    def forward(self, *args, **kwargs):
        return args[0]

    def extra_repr(self) -> str:
        repr = super(QuantDequantNone, self).extra_repr()
        return f"{repr}, doesn't quantize nor dequantize"


class QuantInput(QuantDequantBase):
    def __init__(self, scale_inv, lp_dtype, hp_dtype, *args, **kwargs):
        super(QuantInput, self).__init__(lp_dtype, hp_dtype, *args, **kwargs)
        scale_inv = scale_inv.unsqueeze(1) if (scale_inv.numel() > 1 and not self.use_qdq) else scale_inv
        self.register_scale("scale_inv", scale_inv, self.scale_format)
        if self.use_qdq:
            self.register_scale("scale", 1 / self.scale_inv, self.scale_format)
            op_type = OP_TYPE.QUANT_PC if self.scale_format == ScaleFormat.CONST and self.scale.numel() > 1 else OP_TYPE.QUANT
        else:
            op_type = OP_TYPE.CAST_TO_FP8
        self.quantize_op = get_quantized_func_wrapper(op_type, self.scale_format)

    def forward(self, x):
        return self.quantize_op(x, self.scale_inv, False, False, self.lp_dtype)

    def forward_qdq(self, x):
        return self.quantize_op(
                x,
                scale=self.scale,
                zero_point=self.zero_point,
                axis=0,
                quant_min=self.quant_min,
                quant_max=self.quant_max,
                dtype=self.lp_dtype,
            )

    def extra_repr(self) -> str:
        repr = super(QuantInput, self).extra_repr()
        dtype = get_scale_dtype(self.scale_inv)
        return f"{repr}, scale_inv dtype={dtype}"


class QuantDynamicInput(QuantDequantBase):
    def __init__(self, input_scales_creator, lp_dtype, hp_dtype, *args, **kwargs):
        super(QuantDynamicInput, self).__init__(lp_dtype, hp_dtype, *args, **kwargs)
        self.input_scales_creator = input_scales_creator

        self.cast_to_op = get_quantized_func_wrapper(OP_TYPE.CAST_TO_FP8, self.scale_format)

    def calculate_scales(self, x):
        scale = self.input_scales_creator.calc_scales(x, QuantTensorType.DYNAMIC)
        scale_inv = self.input_scales_creator.invert_scales(scale)
        return scale, scale_inv

    def forward(self, x):
        scale, scale_inv = self.calculate_scales(x)

        ret = self.cast_to_op(x, scale_inv, False, False, self.lp_dtype)

        return ret, scale

    #TODO [SW-224609]: implement forward qdq

    def extra_repr(self) -> str:
        repr = super(QuantDynamicInput, self).extra_repr()
        return f"{repr} input_scales_creator={self.input_scales_creator}"


class DequantOutput(QuantDequantBase):
    def __init__(self, scale, lp_dtype, hp_dtype, *args, **kwargs):
        super(DequantOutput, self).__init__(lp_dtype, hp_dtype, *args, **kwargs)
        self.register_scale("scale", scale, self.scale_format)
        if self.use_qdq:
            op_type = OP_TYPE.DEQUANT_PC if self.scale_format == ScaleFormat.CONST and self.scale.numel() > 1 else OP_TYPE.DEQUANT
        else:
            op_type = OP_TYPE.CAST_FROM_FP8
        self.dequantize_op = get_quantized_func_wrapper(op_type, self.scale_format)

    def forward(self, x):
        return self.dequantize_op(x, self.scale, self.hp_dtype)

    def forward_qdq(self, x):
        return self.dequantize_op(
                x,
                scale= self.scale,
                zero_point=self.zero_point,
                axis=1,
                quant_min=self.quant_min,
                quant_max=self.quant_max,
                dtype=self.lp_dtype,
                out_dtype=self.hp_dtype,
            )

    def extra_repr(self) -> str:
        repr = super(DequantOutput, self).extra_repr()
        dtype = get_scale_dtype(self.scale)
        return f"{repr}, scale dtype={dtype}"


class QuantDequant(QuantDequantBase):
    def __init__(self, scale_inv, lp_dtype, hp_dtype, *args, **kwargs):
        super(QuantDequant, self).__init__(lp_dtype, hp_dtype, *args, **kwargs)
        self.register_scale("scale_inv", scale_inv, self.scale_format)
        self.register_scale("scale", 1 / scale_inv, self.scale_format)
        self.quantize_op = (
           get_quantized_func_wrapper(OP_TYPE.QUANT, self.scale_format)
           if self.use_qdq
           else get_quantized_func_wrapper(OP_TYPE.CAST_TO_FP8, self.scale_format)
        )
        self.dequantize_op = (
           get_quantized_func_wrapper(OP_TYPE.DEQUANT, self.scale_format)
           if self.use_qdq
           else get_quantized_func_wrapper(OP_TYPE.CAST_FROM_FP8, self.scale_format)
        )

    def forward(self, x, *args, **kwargs):
        y = self.quantize_op(x, self.scale_inv, False, False, self.lp_dtype)
        # mark_step is needed so fuser won't remove 2 consecutive casts.
        # will be removed once SW-196431 is implemented
        # Call cur_accelerator.synchronize() which will call mark_step() as well
        cur_accelerator.synchronize()
        z = self.dequantize_op(y, self.scale, self.hp_dtype)
        cur_accelerator.synchronize()
        return z

    def forward_qdq(self, x, *args, **kwargs):
        y = self.quantize_op(
            x,
            scale=self.scale,
            zero_point=self.zero_point,
            quant_min=self.quant_min,
            quant_max=self.quant_max,
            dtype=self.lp_dtype,
        )
        z = self.dequantize_op(
            y,
            scale=self.scale,
            zero_point=self.zero_point,
            quant_min=self.quant_min,
            quant_max=self.quant_max,
            dtype=self.lp_dtype,
            out_dtype=self.hp_dtype,
        )
        return z

    def extra_repr(self) -> str:
        repr = super(QuantDequant, self).extra_repr()
        return f"{repr}, Quantize, and then dequantize"
